"""Defines interfaces for registering functions."""
import inspect
import json
import traceback
import typing
import websockets

import iterm2.capabilities
import iterm2.notifications
import iterm2.rpc


async def generic_handle_rpc(coro, connection, notif):
    """Dispatches an inbound RPC message."""
    rpc_notif = notif.server_originated_rpc_notification
    params = {}
    successful = False
    # pylint: disable=broad-except
    try:
        for arg in rpc_notif.rpc.arguments:
            name = arg.name
            if arg.HasField("json_value"):
                # NOTE: This can throw an exception if there are control
                # characters or other nasties.
                value = json.loads(arg.json_value)
                params[name] = value
            else:
                params[name] = None
        result = await coro(**params)
        successful = True
    except KeyboardInterrupt as exception:
        raise exception
    except websockets.exceptions.ConnectionClosed as exception:
        raise exception
    except Exception as exception:
        stack_trace = traceback.format_exc()
        exception = {"reason": repr(exception), "traceback": stack_trace}
        await iterm2.rpc.async_send_rpc_result(
            connection, rpc_notif.request_id, True, exception)

    if successful:
        await iterm2.rpc.async_send_rpc_result(
            connection, rpc_notif.request_id, False, result)

class Reference:  # pylint: disable=too-few-public-methods
    """Defines a reference to a variable for use in the @RPC decorator.

    .. seealso::
        * :func:`~iterm2.registration.RPC`.
        * Example ":ref:`mousemode_example`"
        * Example ":ref:`statusbar_example`"
    """
    def __init__(self, name):
        self.name = name


def RPC(func):  # pylint: disable=invalid-name
    """
    A decorator that adds an `async_register` value to the coroutine it
    decorates. `async_register` is a coroutine that will register the function
    as an RPC.

    An RPC is a function in a script that iTerm2 can invoke in response to some
    action, such a a keypress or a trigger. You use this decorator to register
    a coroutine as an RPC.

    Every RPC must have a unique signature. The signature is composed of two
    parts: first, the name, which comes from the name of the coroutine being
    decorated; second, the names of its arguments. The order of arguments is
    not important.

    The decorated coroutine will have a `async_register` value that you must
    call to complete the registration. `async_register` takes one required
    argument, the :class:`~iterm2.connection.Connection`. It also takes one
    optional argument, which is a timeout. The `timeout` is a value in seconds.
    If not given, the default timeout will be used. When waiting for an RPC to
    return, iTerm2 will stop waiting for the RPC after the timeout elapses.

    Do not use default values for arguments in your decorated coroutine, with
    one exception: a special kind of default value of type
    :class:`iterm2.Reference`. It names a variable that is visible in the
    context of the invocation. It will be transformed to the current value of
    that variable. This is the only way to get information about the current
    context. For example, a value of `iterm2.Reference("id")` will give you the
    session ID of the context where the RPC was invoked. If the RPC is run from
    a keyboard shortcut, that is the ID of the session that had keyboard focus
    at the time of invocation.

    .. seealso::
        * Example ":ref:`badgetitle_example`"
        * Example ":ref:`blending_example`"
        * Example ":ref:`close_to_the_right_example`"
        * Example ":ref:`cls_example`"
        * Example ":ref:`jsonpretty_example`"
        * Example ":ref:`movetab_example`"

    That's complicated, but an example will make it clearer:

    Example:

      .. code-block:: python

          app = await iterm2.async_get_app(connection)

          @iterm2.RPC
          async def split_current_session_n_times(
                  session_id=iterm2.Reference("id"), n=1):
              session = app.get_session_by_id(session_id)
              for i in range(n):
                  await session.async_split_pane()

          # Remember to call async_register!
          await split_current_session_n_times.async_register(connection)
    """
    async def async_register(connection, timeout=None):
        signature = inspect.signature(func)
        defaults = {}
        for key, value in signature.parameters.items():
            if isinstance(value.default, Reference):
                defaults[key] = value.default.name

        async def handle_rpc(connection, notif):
            await generic_handle_rpc(func, connection, notif)

        func.rpc_token = (
            await iterm2.notifications.
            async_subscribe_to_server_originated_rpc_notification(
                connection,
                handle_rpc,
                func.__name__,
                signature.parameters.keys(),
                timeout,
                defaults,
                iterm2.notifications.RPC_ROLE_GENERIC))
        func.rpc_connection = connection

    func.async_register = async_register
    return func

def ContextMenuProviderRPC(func):  # pylint: disable=invalid-name
    """
    A decorator that prepares a function for registration as a context menu
    provider. Similar to :func:`~iterm2.registration.RPC`.

    A context menu provider is a function that gets called when the user selects
    an item in a context menu. The context menu is the menu that appears when you
    right-click in a terminal window.

    The return value is ignored.

    :param connection: :class:`Connection`
    :param display_name: Name that appears in the context menu.
    :param unique_identifier: Globally unique identifier for this provider
        (e.g., "com.example.my-provider")
    :param timeout: Max number of seconds to wait, or None to use the default.
    """
    async def async_register(
            connection: iterm2.connection.Connection,
            display_name: str,
            unique_identifier: str,
            timeout: typing.Optional[float] = None):
        assert unique_identifier
        iterm2.capabilities.check_supports_context_menu_provider(connection)
        signature = inspect.signature(func)
        defaults = {}
        for key, value in signature.parameters.items():
            if isinstance(value.default, Reference):
                defaults[key] = value.default.name

        async def handle_rpc(connection, notif):
            await generic_handle_rpc(func, connection, notif)

        func.rpc_token = (
            await iterm2.notifications.
            async_subscribe_to_server_originated_rpc_notification(
                connection=connection,
                callback=handle_rpc,
                name=func.__name__,
                arguments=signature.parameters.keys(),
                timeout_seconds=timeout,
                defaults=defaults,
                role=iterm2.notifications.RPC_ROLE_CONTEXT_MENU,
                context_menu_display_name=display_name,
                context_menu_unique_id=unique_identifier))
        func.rpc_connection = connection

    func.async_register = async_register
    return func

def TitleProviderRPC(func):  # pylint: disable=invalid-name
    """
    A decorator that prepares a function for registration as a session title
    provider. Similar to :func:`~iterm2.registration.RPC`.

    A session title provider is a function that gets called to compute the
    title of a session. It may be called frequently, whenever the session title
    is deemed to need recomputation. Once registered, it appears as an option
    in the list of title settings in preferences.

    It is called when any of its inputs change. This will only happen if one or
    more of the inputs are :func:`~iterm2.registration.Reference` references to
    variables in the session context.

    It must return a string.

    Note that the `async_register` function is different than in the
    :func:`~iterm2.registration.RPC` decorator: it takes three arguments. The
    first is the :class:`~iterm2.connection.Connection`. The second is a
    "display name", which is the string to show in preferences that the user
    may select to use this title provider. The third is a string identifier,
    which must be unique among all title providers. The identifier should be a
    reverse DNS name, like `com.example.my-title-provider`. As long as the
    identifier remains the same from one version to the next, the display name
    and function signature may change.

    .. seealso:: Example ":ref:`georges_title_example`"

    Example:

      .. code-block:: python

          @iterm2.TitleProviderRPC
          async def upper_case_title(auto_name=iterm2.Reference("autoName?")):
              if not auto_name:
                  return ""
              return auto_name.upper()

          # Remember to call async_register!
          await upper_case_title.async_register(
                  connection,
                  display_name="Upper-case Title",
                  unique_identifier="com.iterm2.example.title-provider")
    """
    async def async_register(
            connection, display_name, unique_identifier, timeout=None):
        assert unique_identifier
        signature = inspect.signature(func)
        defaults = {}
        for key, value in signature.parameters.items():
            if isinstance(value.default, Reference):
                defaults[key] = value.default.name

        async def handle_rpc(connection, notif):
            await generic_handle_rpc(func, connection, notif)

        func.rpc_token = (
            await iterm2.notifications.
            async_subscribe_to_server_originated_rpc_notification(
                connection=connection,
                callback=handle_rpc,
                name=func.__name__,
                arguments=signature.parameters.keys(),
                timeout_seconds=timeout,
                defaults=defaults,
                role=iterm2.notifications.RPC_ROLE_SESSION_TITLE,
                session_title_display_name=display_name,
                session_title_unique_id=unique_identifier))
        func.rpc_connection = connection

    func.async_register = async_register
    return func


def StatusBarRPC(func):  # pylint: disable=invalid-name
    """
    A decorator (like :func:`~iterm2.registration.RPC`) that registers a custom
    status bar component.

    See :class:`~iterm2.statusbar.StatusBarComponent` for details on what a
    status bar component is.

    The coroutine is called when any of its inputs change. This will only
    happen if one or more of the inputs are
    :func:`~iterm2.registration.Reference` references to variables in the
    session context.

    The coroutine *must* take an argument named `knobs` that will contain a
    dictionary with configuration settings.

    It may return a string or an array of strings. In the case that it returns
    an array, the longest string fitting the available space will be used.

    Note that unlike the other RPC decorators, you use
    :meth:`~iterm2.statusbar.StatusBarComponent.async_register` to register it,
    rather than a register property added to the coroutine.

    .. seealso::
        * Example ":ref:`escindicator_example`"
        * Example ":ref:`jsonpretty_example`"
        * Example ":ref:`mousemode_example`"
        * Example ":ref:`statusbar_example`"

    Example:

      .. code-block:: python

          component = iterm2.StatusBarComponent(
              short_description="Session ID",
              detailed_description="Show the session's identifier",
              knobs=[],
              exemplar="[session ID]",
              update_cadence=None,
              identifier="com.iterm2.example.statusbar-rpc")

          @iterm2.StatusBarRPC
          async def session_id_status_bar_coro(
                  knobs,
                  session_id=iterm2.Reference("id")):
              # This status bar component shows the current session ID, which
              # is useful for debugging scripts.
              return session_id

          @iterm2.RPC
          async def my_status_bar_click_handler(session_id):
              # When you click the status bar it opens a popover with the
              # message "Hello World"
              await component.async_open_popover(
                      session_id,
                      "Hello world",
                      iterm2.Size(200, 200))

          await component.async_register(
                  connection,
                  session_id_status_bar_coro,
                  onclick=my_status_bar_click_handler)
    """
    async def async_register(connection, component, timeout=None):
        signature = inspect.signature(func)
        defaults = {}
        for key, value in signature.parameters.items():
            if isinstance(value.default, Reference):
                defaults[key] = value.default.name

        async def wrapper(**kwargs):
            """handle_rpc->generic_handle_rpc->wrapper->func"""
            # Fix up knobs to not be JSON
            if "knobs" in kwargs:
                knobs_json = kwargs["knobs"]
                kwargs["knobs"] = json.loads(knobs_json)
            return await func(**kwargs)

        async def handle_rpc(connection, notif):
            """This gets run first."""
            await generic_handle_rpc(wrapper, connection, notif)

        func.rpc_token = (
            await iterm2.notifications.
            async_subscribe_to_server_originated_rpc_notification(
                connection=connection,
                callback=handle_rpc,
                name=func.__name__,
                arguments=signature.parameters.keys(),
                timeout_seconds=timeout,
                defaults=defaults,
                role=iterm2.notifications.RPC_ROLE_STATUS_BAR_COMPONENT,
                status_bar_component=component))
        func.rpc_connection = connection

    func.async_register = async_register

    return func



"""Subscribe/unsubscribe from async notifications.

This module provides functions that let you subscribe and unsubscribe from
notifications. iTerm2 posts notifications when some event of interest (for
example, a keystroke) occurs. By subscribing to a notifications your async
callback will be run when the event occurs.
"""
import iterm2.api_pb2
import iterm2.connection
import iterm2.rpc

# pylint: disable=no-member
RPC_ROLE_GENERIC = iterm2.api_pb2.RPCRegistrationRequest.Role.Value("GENERIC")
RPC_ROLE_SESSION_TITLE = iterm2.api_pb2.RPCRegistrationRequest.Role.Value(
        "SESSION_TITLE")
RPC_ROLE_STATUS_BAR_COMPONENT = (
    iterm2.api_pb2.RPCRegistrationRequest.Role.Value("STATUS_BAR_COMPONENT"))
RPC_ROLE_CONTEXT_MENU = iterm2.api_pb2.RPCRegistrationRequest.Role.Value(
        "CONTEXT_MENU")

# pylint: enable=no-member


def _get_handlers():
    """Returns the registered notification handlers.

    :returns: (session, notification_type) -> [coroutine, ...]
    """
    if not hasattr(_get_handlers, 'handlers'):
        _get_handlers.handlers = {}
    return _get_handlers.handlers

# APIs -----------------------------------------------------------------------


class SubscriptionException(Exception):
    """Raised when a subscription attempt fails."""


async def async_unsubscribe(connection, token):
    """
    Unsubscribes from a notification.

    :param connection: A connected :class:`Connection`.
    :param token: The result of a previous subscribe call.
    """
    key, coro = token
    coros = _get_handlers()[key]
    coros.remove(coro)
    if coros:
        _get_handlers()[key] = coros
    else:
        del _get_handlers()[key]
        if len(key) == 2:
            session, notification_type = key
            await _async_subscribe(
                connection,
                False,
                notification_type,
                coro,
                session=session)
        else:
            if key[-1] == iterm2.api_pb2.NOTIFY_ON_VARIABLE_CHANGE:
                scope, identifier, name, notification_type = key
                request = iterm2.api_pb2.VariableMonitorRequest()
                request.name = name
                request.scope = scope
                request.identifier = identifier
                await _async_subscribe(
                    connection,
                    False,
                    notification_type,
                    coro,
                    key=key,
                    variable_monitor_request=request)
            else:
                # The key is custom and there should be code to handle
                # unsubscribing as an elif clause here.
                assert False


async def async_subscribe_to_new_session_notification(connection, callback):
    """
    Registers a callback to be run when a new session is created.

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: an
      :class:`Connection` and iterm2.api_pb2.NewSessionNotification.

    :returns: A token that can be passed to unsubscribe.
    """
    return await _async_subscribe(
        connection, True, iterm2.api_pb2.NOTIFY_ON_NEW_SESSION, callback)


async def async_subscribe_to_keystroke_notification(
        connection, callback, session=None, advanced=False):
    """
    Registers a callback to be run when a key is pressed.

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: an :class:`Connection`
        and iterm2.api_pb2.KeystrokeNotification.
    :param session: The session to monitor, or None.
    :param advanced: Use "advanced" reporting (key up, flags changed)?

    Returns: A token that can be passed to unsubscribe.
    """
    kmr = iterm2.api_pb2.KeystrokeMonitorRequest()
    kmr.advanced = advanced
    return await _async_subscribe(
        connection,
        True,
        iterm2.api_pb2.NOTIFY_ON_KEYSTROKE,
        callback,
        session=session,
        keystroke_monitor_request=kmr)


async def async_filter_keystrokes(
        connection, patterns_to_ignore, session=None):
    """
    Subscribe to a pseudo-notification that causes keystrokes matching patterns
    to be ignored.

    :param connection: A connected :class:`Connection`.
    :param patterns_to_ignore: A list of keystroke patterns that iTerm2 should
        not handle, expecting the script will handle it alone. Objects are
        class :class:`KeystrokePattern`.
    :param session: The session to monitor, or None.

    Returns: A token that can be passed to unsubscribe.
    """
    # pylint: disable=no-member
    kfr = iterm2.api_pb2.KeystrokeFilterRequest()
    kfr.patterns_to_ignore.extend(
        list(map(lambda x: x.to_proto(), patterns_to_ignore)))

    async def callback(connection, notif):  # pylint: disable=unused-argument
        pass

    return await _async_subscribe(
        connection,
        True,
        iterm2.api_pb2.KEYSTROKE_FILTER,
        callback,
        session=session,
        keystroke_filter_request=kfr)


async def async_subscribe_to_screen_update_notification(
        connection, callback, session=None):
    """
    Registers a callback to be run when the screen contents change.

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: an :class:`Connection`
        and iterm2.api_pb2.ScreenUpdateNotification..
    :param session: The session to monitor, or None.

    :returns: A token that can be passed to unsubscribe.
    """
    return await _async_subscribe(
        connection,
        True,
        iterm2.api_pb2.NOTIFY_ON_SCREEN_UPDATE,
        callback,
        session=session)


async def async_subscribe_to_prompt_notification(
        connection, callback, session, modes):
    """
    Registers a callback to be run when a shell prompt is received.

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: an :class:`Connection`
        and iterm2.api_pb2.PromptNotification.
    :param session: The session to monitor, or None.
    :param modes: The modes to subscribe to.

    :returns: A token that can be passed to unsubscribe.
    """
    return await _async_subscribe(
        connection,
        True,
        iterm2.api_pb2.NOTIFY_ON_PROMPT,
        callback,
        session=session,
        prompt_monitor_modes=modes)


async def async_subscribe_to_custom_escape_sequence_notification(connection,
                                                                 callback,
                                                                 session=None):
    """
    Registers a callback to be run when a custom escape sequence is received.

    The escape sequence is OSC 1337 ; Custom=id=<identity>:<payload> ST

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: an :class:`Connection`
        and iterm2.api_pb2.CustomEscapeSequenceNotification.
    :param session: The session to monitor, or None.

    :returns: A token that can be passed to unsubscribe.
    """
    return await _async_subscribe(
        connection,
        True,
        iterm2.api_pb2.NOTIFY_ON_CUSTOM_ESCAPE_SEQUENCE,
        callback,
        session=session)


async def async_subscribe_to_terminate_session_notification(
        connection, callback):
    """
    Registers a callback to be run when a session terminates.

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: an :class:`Connection`
        and iterm2.api_pb2.TerminateSessionNotification.

    :returns: A token that can be passed to unsubscribe.
    """
    return await _async_subscribe(
        connection,
        True,
        iterm2.api_pb2.NOTIFY_ON_TERMINATE_SESSION,
        callback,
        session=None)


async def async_subscribe_to_layout_change_notification(connection, callback):
    """
    Registers a callback to be run when the relationship between sessions,
    tabs, and windows changes.

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: an :class:`Connection`
        and iterm2.api_pb2.LayoutChangedNotification.

    :returns: A token that can be passed to unsubscribe.
    """
    return await _async_subscribe(
        connection,
        True,
        iterm2.api_pb2.NOTIFY_ON_LAYOUT_CHANGE,
        callback,
        session=None)


async def async_subscribe_to_focus_change_notification(connection, callback):
    """
    Registers a callback to be run when focus changes.

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: an :class:`Connection`
        and iterm2.api_pb2.FocusChangedNotification.

    :returns: A token that can be passed to unsubscribe.
    """
    return await _async_subscribe(
        connection,
        True,
        iterm2.api_pb2.NOTIFY_ON_FOCUS_CHANGE,
        callback,
        session=None)


async def async_subscribe_to_broadcast_domains_change_notification(
        connection, callback):
    """
    Registers a callback to be run when the current broadcast domains change.

    See also: :meth:`~iTerm2.App.parse_broadcast_domains`. Pass it
        `notification.broadcast_domains`.

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: an :class:`Connection`
        and iterm2.api_pb2.BroadcastDomainsChangedNotification.
    """
    return await _async_subscribe(
        connection,
        True,
        iterm2.api_pb2.NOTIFY_ON_BROADCAST_CHANGE,
        callback,
        session=None)


# pylint: disable=dangerous-default-value
# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
async def async_subscribe_to_server_originated_rpc_notification(
        connection,
        callback,
        name,
        arguments=[],
        timeout_seconds=5,
        defaults={},
        role=RPC_ROLE_GENERIC,
        session_title_display_name=None,
        session_title_unique_id=None,
        status_bar_component=None,
        context_menu_display_name=None,
        context_menu_unique_id=None):
    """
    Registers a callback to be run when the server wants to invoke an RPC.

    You probably want to use
    :meth:`~iterm2.Registration.async_register_rpc_handler` instead of this.
    It's a much higher level API.

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: an :class:`Connection`
        and iterm2.api_pb2.ServerOriginatedRPCNotification.
    :param timeout_seconds: How long iTerm2 should wait for this function to
        return or `None` to use the default timeout.
    :param defaults: Gives default values. Names correspond to argument names
        in `arguments`. Values are in-scope variables at the callsite.
    :param role: Defines the special purpose of this RPC. If none, use
        `RPC_ROLE_GENERIC`.
    :param session_title_display_name: Used by the `RPC_ROLE_SESSION_TITLE`
        role to give the name of the function to show in preferences.
    :param context_menu_display_name: Used by the `RPC_ROLE_CONTEXT_MENU`
        role to give the name of the menu item.
    :param context_menu_unique_id: Unique ID for the context menu provider.

    :returns: A token that can be passed to unsubscribe.
    """
    # pylint: disable=no-member
    rpc_registration_request = iterm2.api_pb2.RPCRegistrationRequest()
    rpc_registration_request.name = name
    if timeout_seconds is not None:
        rpc_registration_request.timeout = timeout_seconds
    args = []
    for arg_name in arguments:
        arg = iterm2.api_pb2.RPCRegistrationRequest.RPCArgumentSignature()
        arg.name = arg_name
        args.append(arg)
    rpc_registration_request.arguments.extend(args)
    rpc_registration_request.role = role

    if len(defaults) > 0:
        protos = []
        for default_name in defaults:
            assert default_name in arguments, (
                "Name for default '{}' not in arguments {}".format(
                    default_name, arguments))
            path = defaults[default_name]
            argd = iterm2.api_pb2.RPCRegistrationRequest.RPCArgument()
            argd.name = default_name
            argd.path = path
            protos.append(argd)

        rpc_registration_request.defaults.extend(protos)

    if session_title_display_name is not None:
        (rpc_registration_request.session_title_attributes.
         display_name) = session_title_display_name
        assert session_title_unique_id
        rpc_registration_request.session_title_attributes.unique_identifier = (
            session_title_unique_id)
    elif status_bar_component is not None:
        status_bar_component.set_fields_in_proto(
            rpc_registration_request.status_bar_component_attributes)
    elif context_menu_display_name is not None:
        (rpc_registration_request.context_menu_attributes.
                display_name) = context_menu_display_name
        assert context_menu_unique_id
        (rpc_registration_request.context_menu_attributes.
                unique_identifier) = context_menu_unique_id

    return await _async_subscribe(
        connection,
        True,
        iterm2.api_pb2.NOTIFY_ON_SERVER_ORIGINATED_RPC,
        callback,
        rpc_registration_request=rpc_registration_request)
# pylint: enable=dangerous-default-value
# pylint: enable=too-many-arguments
# pylint: enable=too-many-locals


async def async_subscribe_to_variable_change_notification(
        connection, callback, scope, name, identifier):
    """
    Registers a callback to be invoked when a variable changes.

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: an :class:`Connection`
        and iterm2.api_pb2.VariableChangedNotification.
    :param scope: A :class:`VariableScopes` enumerated value.
    :param name: The name of the variable, a string.
    :param identifier: The identifier of the object (window, tab, or session)
        being monitored, or None for app. Sometimes this will be "all" or
        "active".
    """
    # pylint: disable=no-member
    request = iterm2.api_pb2.VariableMonitorRequest()
    request.name = name
    request.scope = scope
    if identifier is not None:
        request.identifier = identifier
    else:
        request.identifier = ""
    key = (request.scope,
           request.identifier,
           request.name,
           iterm2.api_pb2.NOTIFY_ON_VARIABLE_CHANGE)
    return await _async_subscribe(
        connection,
        True,
        iterm2.api_pb2.NOTIFY_ON_VARIABLE_CHANGE,
        callback,
        session=None,
        variable_monitor_request=request,
        key=key)


async def async_subscribe_to_profile_change_notification(
        connection, callback, guid):
    """
    Register a callback to be invoked when a profile changes.

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: a :class:`Connection`
        and iterm2.api_pb2.ProfileChangedNotification.
    :parm guid: The guid to monitor. A string.
    """
    request = iterm2.api_pb2.ProfileChangeRequest()
    request.guid = guid
    key = (guid, iterm2.api_PB2.NOTIFY_ON_PROFILE_CHANGE)
    return await _async_subscribe(
        connection,
        True,
        iterm2.api_PB2.NOTIFY_ON_PROFILE_CHANGE,
        callback,
        session=None,
        profile_change_request=request,
        key=key)


# Private --------------------------------------------------------------------

def _string_rpc_registration_request(rpc):
    """Converts ServerOriginatedRPC or RPCSignature to a string."""
    if rpc is None:
        return None
    args = sorted(map(lambda x: x.name, rpc.arguments))
    return rpc.name + "(" + ",".join(args) + ")"


# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
async def _async_subscribe(
        connection,
        subscribe,
        notification_type,
        callback,
        session=None,
        rpc_registration_request=None,
        keystroke_monitor_request=None,
        variable_monitor_request=None,
        key=None,
        profile_change_request=None,
        prompt_monitor_modes=None,
        keystroke_filter_request=None):
    """Note: session argument is ignored for variable-change notifications."""
    _register_helper_if_needed()
    transformed_session = session if session is not None else "all"

    # Register locally in case iTerm2 responds with an RPC immediately.
    # That is typical when registering a session title provider when a session
    # is already open.
    if subscribe:
        if key:
            _register_notification_handler_impl(key, callback)
        else:
            _register_notification_handler(
                session,
                _string_rpc_registration_request(rpc_registration_request),
                notification_type,
                callback)

    response = await iterm2.rpc.async_notification_request(
        connection,
        subscribe,
        notification_type,
        transformed_session,
        rpc_registration_request,
        keystroke_monitor_request,
        variable_monitor_request,
        profile_change_request,
        prompt_monitor_modes,
        keystroke_filter_request)
    status = response.notification_response.status
    # pylint: disable=no-member
    status_ok = (
        status == iterm2.api_pb2.NotificationResponse.Status.Value("OK"))

    unregister = False
    if subscribe:
        already = (
            status == iterm2.api_pb2.NotificationResponse.Status.Value(
                "ALREADY_SUBSCRIBED"))
        if status_ok or already:
            if key:
                return (key, callback)
            return ((session, notification_type), callback)
        # Uh oh, undo the registration.
        unregister = True
    else:
        # Unsubscribe
        if status_ok:
            unregister = True

    if unregister:
        if key:
            _unregister_notification_handler_impl(key, callback)
        else:
            _unregister_notification_handler(
                session,
                _string_rpc_registration_request(rpc_registration_request),
                notification_type,
                callback)

        if not subscribe and status_ok:
            # Normal code path for unsubscribe
            return

    # pylint: disable=no-member
    raise SubscriptionException(
        iterm2.api_pb2.NotificationResponse.Status.Name(status))
# pylint: enable=too-many-arguments
# pylint: enable=too-many-locals


def _register_helper_if_needed():
    if not hasattr(_register_helper_if_needed, 'haveRegisteredHelper'):
        _register_helper_if_needed.haveRegisteredHelper = True
        iterm2.connection.Connection.register_helper(_async_dispatch_helper)


async def _async_dispatch_helper(connection, message):
    handlers, sub_notification = _get_notification_handlers(message)
    for handler in handlers:
        assert handler is not None
        await handler(connection, sub_notification)

    if (not handlers and
            message.notification.HasField(
                'server_originated_rpc_notification')):
        # If we get an RPC we haven't registered for handle the error because
        # otherwise it has to time out. If you get here there is probably a
        # bug.
        key, _sub_notification = _get_handler_key_from_notification(
            message.notification)
        exception = {"reason": "No such function: {} (key was {})".format(
            _string_rpc_registration_request(
                message.notification.server_originated_rpc_notification.rpc),
            key)}
        await iterm2.rpc.async_send_rpc_result(
            connection,
            (message.notification.server_originated_rpc_notification.
             request_id),
            True,
            exception)

    return bool(handlers)


def _get_all_sessions_handler_keys_from_notification(notification):
    """Returns keys into _get_handlers() for a notification for the caller may
    have registered for all sessions/windows."""
    if notification.HasField('variable_changed_notification'):
        return [(iterm2.api_pb2.VariableScope.Value("SESSION"),
                 "all",
                 notification.variable_changed_notification.name,
                 iterm2.api_pb2.NOTIFY_ON_VARIABLE_CHANGE),

                (iterm2.api_pb2.VariableScope.Value("WINDOW"),
                 "all",
                 notification.variable_changed_notification.name,
                 iterm2.api_pb2.NOTIFY_ON_VARIABLE_CHANGE)]
    # Convert a session-specific key into an all-sessions key.
    standard_key, _ = _get_handler_key_from_notification(notification)
    return [(None, standard_key[1])]


# pylint: disable=too-many-branches
def _get_handler_key_from_notification(notification):
    key = None

    if notification.HasField('keystroke_notification'):
        key = (notification.keystroke_notification.session,
               iterm2.api_pb2.NOTIFY_ON_KEYSTROKE)
        notification = notification.keystroke_notification
    elif notification.HasField('screen_update_notification'):
        key = (notification.screen_update_notification.session,
               iterm2.api_pb2.NOTIFY_ON_SCREEN_UPDATE)
        notification = notification.screen_update_notification
    elif notification.HasField('prompt_notification'):
        key = (notification.prompt_notification.session,
               iterm2.api_pb2.NOTIFY_ON_PROMPT)
        notification = notification.prompt_notification
    elif notification.HasField('location_change_notification'):
        key = (notification.location_change_notification.session,
               iterm2.api_pb2.NOTIFY_ON_LOCATION_CHANGE)
        notification = notification.location_change_notification
    elif notification.HasField('custom_escape_sequence_notification'):
        key = (notification.custom_escape_sequence_notification.session,
               iterm2.api_pb2.NOTIFY_ON_CUSTOM_ESCAPE_SEQUENCE)
        notification = notification.custom_escape_sequence_notification
    elif notification.HasField('new_session_notification'):
        key = (None, iterm2.api_pb2.NOTIFY_ON_NEW_SESSION)
        notification = notification.new_session_notification
    elif notification.HasField('terminate_session_notification'):
        key = (None, iterm2.api_pb2.NOTIFY_ON_TERMINATE_SESSION)
        notification = notification.terminate_session_notification
    elif notification.HasField('layout_changed_notification'):
        key = (None, iterm2.api_pb2.NOTIFY_ON_LAYOUT_CHANGE)
        notification = notification.layout_changed_notification
    elif notification.HasField('focus_changed_notification'):
        key = (None, iterm2.api_pb2.NOTIFY_ON_FOCUS_CHANGE)
        notification = notification.focus_changed_notification
    elif notification.HasField('server_originated_rpc_notification'):
        key = (None,
               iterm2.api_pb2.NOTIFY_ON_SERVER_ORIGINATED_RPC,
               _string_rpc_registration_request(
                   notification.server_originated_rpc_notification.rpc))
    elif notification.HasField('broadcast_domains_changed'):
        key = (None, iterm2.api_pb2.NOTIFY_ON_BROADCAST_CHANGE)
    elif notification.HasField('variable_changed_notification'):
        key = (notification.variable_changed_notification.scope,
               notification.variable_changed_notification.identifier,
               notification.variable_changed_notification.name,
               iterm2.api_pb2.NOTIFY_ON_VARIABLE_CHANGE)
        notification = notification.variable_changed_notification
    elif notification.HasField('profile_changed_notification'):
        key = (notification.profile_changed_notification.guid,
               iterm2.api_pb2.NOTIFY_ON_PROFILE_CHANGE)
    return key, notification
# pylint: enable=too-many-branches


def _get_notification_handlers(message):
    key, sub_notification = _get_handler_key_from_notification(
        message.notification)
    if key is None:
        return ([], None)

    # These keys catch "all"-style subscriptions.
    all_objects_key = _get_all_sessions_handler_keys_from_notification(
        message.notification)

    if key in _get_handlers():
        return (_get_handlers()[key], sub_notification)
    handlers = _get_handlers()
    for all_objects_key in all_objects_key:
        if all_objects_key in handlers:
            return (handlers[all_objects_key], sub_notification)
    return ([], None)


def _register_notification_handler(
        session, rpc_registration_request, notification_type, coro):
    assert coro is not None

    if rpc_registration_request is None:
        key = (session, notification_type)
    else:
        key = (session, notification_type, rpc_registration_request)
    _register_notification_handler_impl(key, coro)


def _register_notification_handler_impl(key, coro):
    if key in _get_handlers():
        _get_handlers()[key].append(coro)
    else:
        _get_handlers()[key] = [coro]


def _unregister_notification_handler(
        session, rpc_registration_request, notification_type, coro):
    assert coro is not None

    if rpc_registration_request is None:
        key = (session, notification_type)
    else:
        key = (session, notification_type, rpc_registration_request)
    _unregister_notification_handler_impl(key, coro)


def _unregister_notification_handler_impl(key, coro):
    if key in _get_handlers():
        if coro in _get_handlers()[key]:
            _get_handlers()[key].remove(coro)



"""Provides interfaces for interacting with selected text regions."""
import enum
import itertools
from operator import itemgetter
import typing

import iterm2.api_pb2
import iterm2.connection
import iterm2.screen
import iterm2.util


class SelectionMode(enum.Enum):
    """Enumerated list of modes for selecting text."""
    CHARACTER = 0  #: character-by-character selection
    WORD = 1  #: word-by-word selection
    LINE = 2  #: row-by-row selection
    SMART = 3  #: smart selection
    BOX = 4  #: rectangular region
    WHOLE_LINE = 5  #: entire wrapped lines, which could occupy many rows

    @staticmethod
    def from_proto_value(value):
        """Creates a SelectionMode from a protobuf enum value."""
        # pylint: disable=global-statement
        global MODE_MAP
        return MODE_MAP[value]

    @staticmethod
    def to_proto_value(value):
        """Converts self into a protobuf enum value."""
        # pylint: disable=global-statement
        global INVERSE_MODE_MAP
        return INVERSE_MODE_MAP[value]


class SubSelection:
    """Describes a continguous block of selected characters.

    :param windowed_coord_range: A :class:`~iterm2.util.WindowedCoordRange`
        describing the range.
    :param mode: A :class:`SelectionMode` describing how the selection is
        interpreted and extended.
    :param connected: If true, no newline exists between this and the next
        sub-selection.
    """
    def __init__(
            self,
            windowed_coord_range: iterm2.util.WindowedCoordRange,
            mode: SelectionMode,
            connected: bool):
        self.__windowed_coord_range = windowed_coord_range
        self.__mode = mode
        self.__connected = connected

    # pylint: disable=invalid-name
    @property
    def windowedCoordRange(self) -> iterm2.util.WindowedCoordRange:
        """Deprecated in favor of windowed_coord_range"""
        return self.windowed_coord_range
    # pylint: enable=invalid-name

    @property
    def windowed_coord_range(self) -> iterm2.util.WindowedCoordRange:
        """Returns the coordinate range spanned by this sub-selection."""
        return self.__windowed_coord_range

    @property
    def mode(self) -> SelectionMode:
        """Returns the mode of this sub-selection."""
        return self.__mode

    @property
    def proto(self):
        """Creates a protobuf for this sub-selection."""
        # pylint: disable=no-member
        sub_selection = iterm2.api_pb2.SubSelection()
        sub_selection.windowed_coord_range.CopyFrom(self.__windowed_coord_range.proto)
        sub_selection.selection_mode = SelectionMode.to_proto_value(self.__mode)
        return sub_selection

    @property
    def connected(self) -> bool:
        """
        Returns whether the subselection is connected to the next
        subselection.
        """
        return self.__connected

    async def async_get_string(
            self,
            connection: iterm2.connection.Connection,
            session_id: str) -> str:
        """Gets the text belonging to this subselection.

        :param connection: The connection to iTerm2.
        :param session_id: The ID of the session for which to look up the
            selected text.
        """
        result = await iterm2.rpc.async_get_screen_contents(
            connection,
            session_id,
            self.__windowed_coord_range)
        # pylint: disable=no-member
        if (result.get_buffer_response.status == iterm2.
                api_pb2.GetBufferResponse.Status.Value("OK")):
            screen_contents = iterm2.screen.ScreenContents(
                result.get_buffer_response)
            built_string = ""
            i = 0
            while i < screen_contents.number_of_lines:
                line = screen_contents.line(i)
                i += 1
                built_string += line.string
                if line.hard_eol:
                    built_string += "\n"
            return built_string
        raise iterm2.rpc.RPCException(
            iterm2.api_pb2.GetBufferResponse.Status.Name(
                result.get_buffer_response.status))

    # pylint: disable=invalid-name
    def enumerate_ranges(self, callback):
        """Invoke callback for each selected range."""
        if self.__windowed_coord_range.hasWindow:
            right = self.__windowed_coord_range.right
            start_x = self.__windowed_coord_range.start.x
            y = self.__windowed_coord_range.coordRange.start.y
            while y < self.__windowed_coord_range.coordRange.end.y:
                callback(iterm2.util.CoordRange(
                    iterm2.util.Point(start_x, y),
                    iterm2.util.Point(right, y)))
                start_x = self.__windowed_coord_range.left

                y += 1

            callback(iterm2.util.CoordRange(
                iterm2.util.Point(
                    start_x,
                    self.__windowed_coord_range.coordRange.end.y),
                iterm2.util.Point(
                    self.__windowed_coord_range.end.x,
                    self.__windowed_coord_range.coordRange.end.y)))
        else:
            callback(self.__windowed_coord_range.coordRange, self)
    # pylint: enable=invalid-name


class Selection:
    """
    A collection of :class:`SubSelection` objects, describing all the
    selections in a session.

    :param sub_selections: An array of :class:`SubSelection` objects.
    """
    def __init__(self, sub_selections: typing.List[SubSelection]):
        self.__sub_selections = sub_selections

    # pylint: disable=invalid-name
    @property
    def subSelections(self) -> typing.List[SubSelection]:
        """Deprecated in favore of sub_selections."""
        return self.sub_selections
    # pylint: enable=invalid-name

    @property
    def sub_selections(self) -> typing.List[SubSelection]:
        """Returns the set of subselections."""
        return self.__sub_selections

    async def _async_get_content_in_range(
            self, connection, session_id, coord_range):
        """Returns the string in the given range."""
        result = await iterm2.rpc.async_get_screen_contents(
            connection,
            session_id,
            coord_range)
        # pylint: disable=no-member
        if (result.get_buffer_response.status == iterm2.
                api_pb2.GetBufferResponse.Status.Value("OK")):
            screen_contents = iterm2.screen.ScreenContents(
                result.get_buffer_response)
            built_string = ""
            i = 0
            while i < screen_contents.number_of_lines:
                line = screen_contents.line(i)
                i += 1
                built_string += line.string
                if line.hard_eol:
                    built_string += "\n"
            return built_string

    async def async_get_string(
            self,
            connection: iterm2.connection.Connection,
            session_id: str,
            width: int) -> str:
        """Returns the selected text.

        :param connection: The connection to iTerm2.
        :param session_id: The ID of the session for which to look up the
            selected text.
        :param width: The width (number of columns) of the session.
        """
        if len(self.__sub_selections) == 1:
            return await self.__sub_selections[0].async_get_string(
                connection, session_id)

        result = ""

        async def handle_range(coord_range, eol):
            content = await self._async_get_content_in_range(
                connection, session_id, coord_range)
            nonlocal result
            result += content
            if eol and not content.endswith("\n"):
                result += "\n"

        await self.async_enumerate_ranges(width, handle_range)
        return result

    async def async_enumerate_ranges(self, width, callback):
        """
        Gets the text belonging to each subselection and concatenates them with
        newlines.

        :param width: The width of the session
        :param callback: A function to call for each range, taking a
            WindowedCoordRange and a boolean which is true if there's a hard
            EOL at the end of the range.

        :returns: A string with the selected text.
        """
        # pylint: disable=too-many-locals
        if len(self.__sub_selections) == 0:
            return

        # Ranges ending at connectors don't get a newline following.
        connectors = set()
        indexes = set()
        for outer in self.__sub_selections:
            if outer.connected:
                the_position = (
                    outer.windowed_coord_range.coordRange.end.x +
                    outer.windowed_coord_range.coordRange.end.y * width)
                connectors |= {the_position}

            the_range = iterm2.util.Range(0, 0)

            # It's OK to disable it because handle_range does not escape the
            # loop where outer iterates subselections.
            # pylint: disable=cell-var-from-loop
            def handle_range(outer_range):
                # pylint: disable=invalid-name
                nonlocal indexes
                nonlocal the_range
                nonlocal connectors
                the_range = iterm2.util.Range(
                    outer_range.start.x + outer_range.start.y * width,
                    outer_range.length(width))

                indexes_to_add = the_range.toSet
                indexes_to_remove = set()

                def f(values):
                    i, x = values
                    return i-x

                # Iterate contiguous ranges
                indexes_of_interest = indexes.intersection(indexes_to_add)
                # pylint: disable=unused-variable
                for k, g in itertools.groupby(
                        enumerate(sorted(indexes_of_interest)), f):
                    # range exists in both indexes and the_range
                    values_in_range = map(itemgetter(1), g)
                    indexes_to_remove |= set(values_in_range)
                    indexes_to_add -= set(values_in_range)
                indexes -= indexes_to_remove
                indexes |= indexes_to_add

                # In multipart windowed ranges, add connectors for the endpoint
                # of all but the last # range. Each enumerated range is on its
                # own line.
                if (outer.windowed_coord_range.hasWindow and
                        outer_range.end == outer.windowed_coord_range.
                        coordRange.end and
                        the_range.length > 0):
                    connectors |= {the_range.max}

            outer.enumerate_ranges(handle_range)

        # The ranges may be out of order so put them in an array and then sort
        # it.
        all_ranges = []

        # pylint: disable=invalid-name
        def f(values):
            i, x = values
            return i - x
        # pylint: enable=invalid-name

        # Iterate contiguous ranges
        # pylint: disable=unused-variable
        for k, range_group in itertools.groupby(enumerate(sorted(indexes)), f):
            # range exists in both indexes and the_range
            values_in_range = list(map(itemgetter(1), range_group))
            coord_range = iterm2.util.CoordRange(
                iterm2.util.Point(
                    min(values_in_range) % width,
                    min(values_in_range) // width),
                iterm2.util.Point(
                    max(values_in_range) % width,
                    max(values_in_range) // width))
            all_ranges.append(coord_range)

        def range_key(a_range):
            return a_range.start.y * width + a_range.start.x

        sorted_ranges = sorted(all_ranges, key=range_key)
        for idx, the_range in enumerate(sorted_ranges):
            end_index = (
                the_range.start.x +
                the_range.start.y * width +
                the_range.length(width))
            eol = (end_index not in connectors) and idx + 1 < len(
                sorted_ranges)
            the_range.end.x += 1
            stop = await callback(
                iterm2.util.WindowedCoordRange(the_range), eol)
            if stop:
                break


MODE_MAP = {
    iterm2.api_pb2.SelectionMode.Value("CHARACTER"):
        SelectionMode.CHARACTER,
    iterm2.api_pb2.SelectionMode.Value("WORD"): SelectionMode.WORD,
    iterm2.api_pb2.SelectionMode.Value("LINE"): SelectionMode.LINE,
    iterm2.api_pb2.SelectionMode.Value("SMART"): SelectionMode.SMART,
    iterm2.api_pb2.SelectionMode.Value("BOX"): SelectionMode.BOX,
    iterm2.api_pb2.SelectionMode.Value("WHOLE_LINE"):
        SelectionMode.WHOLE_LINE}

INVERSE_MODE_MAP = {v: k for k, v in MODE_MAP.items()}



"""Shared classes for representing color and related concepts."""

import AppKit
import base64
import enum
import json
import typing


class ColorSpace(enum.Enum):
    """Describes the color space of a :class:`Color`."""
    SRGB = "sRGB"  #: SRGB color space
    CALIBRATED = "Calibrated"  #: Device color space


# pylint: disable=too-many-instance-attributes
class Color:
    """Describes a color.

      :param r: Red, in 0-255
      :param g: Green, in 0-255
      :param b: Blue, in 0-255
      :param a: Alpha, in 0-255
      :param color_space: The color space. Only sRGB is supported currently.
      """
    # pylint: disable=too-many-arguments
    def __init__(
            self,
            r: int = 0,
            g: int = 0,
            b: int = 0,
            a: int = 255,
            color_space: ColorSpace = ColorSpace.SRGB):
        """Create a color."""
        self.__red = r
        self.__green = g
        self.__blue = b
        self.__alpha = a
        self.__color_space = color_space

    @staticmethod
    def from_trigger(s: str) -> typing.Optional['Color']:
        """Decodes a color as encoded in a trigger."""
        color = Color.from_hex(s)
        if color is not None:
            return color
        return Color.from_cocoa(s)

    @staticmethod
    def from_hex(s: str) -> typing.Optional['Color']:
        """Decodes a hex-encoded color like #aabbcc"""
        if s[0] != '#' or len(s) != 7:
            return None
        red = int(s[1:3], 16)
        green = int(s[3:5], 16)
        blue = int(s[5:7], 16)
        return Color(red, green, blue, 255)

    @staticmethod
    def from_cocoa(b: str) -> typing.Optional['Color']:
        """Decodes a NSKeyedArchiver-encoded color."""
        data = base64.b64decode(b)
        nscolor = AppKit.NSColor.alloc().initWithCoder_(AppKit.NSKeyedUnarchiver.alloc().initForReadingWithData_(data))
        return Color(
             round(nscolor.redComponent() * 255),
             round(nscolor.greenComponent() * 255),
             round(nscolor.blueComponent() * 255),
             round(nscolor.alphaComponent() * 255))

    @staticmethod
    def from_legacy_trigger(s: str) -> ('Color', 'Color'):
        i = int(str)

        black = AppKit.NSColor.blackColor
        blue = AppKit.NSColor.blueColor
        brown = AppKit.NSColor.brownColor
        cyan = AppKit.NSColor.cyanColor
        darkgray = AppKit.NSColor.darkGrayColor
        gray = AppKit.NSColor.grayColor
        green = AppKit.NSColor.greenColor
        lightgray = AppKit.NSColor.lightGrayColor
        magenta = AppKit.NSColor.magentaColor
        none = lambda: None
        orange = AppKit.NSColor.orangeColor
        purple = AppKit.NSColor.purpleColor
        red = AppKit.NSColor.redColor
        white = AppKit.NSColor.whiteColor
        yellow = AppKit.NSColor.yellowColor

        table = {
            0: (yellow, black),
            1: (black, yellow),
            2: (white, red),
            3: (red, white),
            4: (black, orange),
            5: (orange, black),
            6: (black, purple),
            7: (purple, black),

            1000: (black, none),
            1001: (darkgray, none),
            1002: (lightgray, none),
            1003: (white, none),
            1004: (gray, none),
            1005: (red, none),
            1006: (green, none),
            1007: (blue, none),
            1008: (cyan, none),
            1009: (yellow, none),
            1010: (magenta, none),
            1011: (orange, none),
            1012: (purple, none),
            1013: (brown, none),

            2000: (none, black),
            2001: (none, darkgray),
            2002: (none, lightgray),
            2003: (none, white),
            2004: (none, gray),
            2005: (none, red),
            2006: (none, green),
            2007: (none, blue),
            2008: (none, cyan),
            2009: (none, yellow),
            2010: (none, magenta),
            2011: (none, orange),
            2012: (none, purple),
            2013: (none, brown),
        }

        (text, background) = table[i]
        return (text(), background())


    def __repr__(self):
        return "({},{},{},{} {})".format(
            round(self.red),
            round(self.green),
            round(self.blue),
            round(self.alpha),
            self.color_space)

    @property
    def red(self) -> int:
        """The color's red component."""
        return self.__red

    @red.setter
    def red(self, value: int):
        """Sets the color's red component."""
        self.__red = value

    @property
    def green(self) -> int:
        """The color's green component."""
        return self.__green

    @green.setter
    def green(self, value: int):
        """Sets the color's green component."""
        self.__green = value

    @property
    def blue(self) -> int:
        """The color's blue component."""
        return self.__blue

    @blue.setter
    def blue(self, value: int):
        """Sets the color's blue component."""
        self.__blue = value

    @property
    def alpha(self) -> int:
        """The color's alpha component."""
        return self.__alpha

    @alpha.setter
    def alpha(self, value: int):
        """Sets the color's alpha component."""
        self.__alpha = value

    @property
    def color_space(self) -> ColorSpace:
        """The color's color space."""
        return self.__color_space

    @color_space.setter
    def color_space(self, value: ColorSpace):
        """Sets the color's scolor space."""
        self.__color_space = value

    def get_dict(self):
        """Returns a dictionary representation of this color.

        Suitable for conversion to a JSON object to pass to iTerm2."""
        return {
            "Red Component": self.red / 255.0,
            "Green Component": self.green / 255.0,
            "Blue Component": self.blue / 255.0,
            "Alpha Component": self.alpha / 255.0,
            "Color Space": self.color_space.value
            }

    def from_dict(self, input_dict):
        """Updates the color from the dictionary's contents."""
        self.red = float(input_dict["Red Component"]) * 255
        self.green = float(input_dict["Green Component"]) * 255
        self.blue = float(input_dict["Blue Component"]) * 255
        if "Alpha Component" in input_dict:
            self.alpha = float(input_dict["Alpha Component"]) * 255
        else:
            self.alpha = 255
        if "Color Space" in input_dict:
            self.color_space = ColorSpace(input_dict["Color Space"])
        else:
            # This is the default because it is what profiles use by default.
            self.color_space = ColorSpace.CALIBRATED

    @property
    def json(self):
        """Returns a JSON representation of this color."""
        return json.dumps(self.get_dict())

    @property
    def hex(self):
        """Returns a #rrggbb representation of this color. Assumes srgb colorspace."""
        two_digit_hex = '02x'
        return "#" + format(self.red, two_digit_hex) + format(self.green, two_digit_hex) + format(self.blue, two_digit_hex)



"""Provides functions to register toolbelt webivew tools."""
import iterm2.api_pb2
import iterm2.connection
import iterm2.rpc


async def async_register_web_view_tool(
        connection: iterm2.connection.Connection,
        display_name: str,
        identifier: str,
        reveal_if_already_registered: bool,
        url: str) -> None:
    """
    Registers a toolbelt tool that shows a webview.

    :param connection: The connection to iTerm2.
    :param display_name: The name of the tool. User-visible.
    :param identifier: A unique ID for this tool. Only one tool with a given
        identifier may be registered at a time.
    :param reveal_if_already_registered: If `True`, shows the tool on a
        duplicate registration attempt.
    :param url: The URL to show in the webview.

    :throws: :class:`~iterm2.RPCException` if something goes wrong

    .. seealso:: Example ":ref:`targeted_input_example`"
    """
    result = await iterm2.rpc.async_register_web_view_tool(
        connection,
        display_name,
        identifier,
        reveal_if_already_registered,
        url)
    status = result.register_tool_response.status
    # pylint: disable=no-member
    if status == iterm2.api_pb2.RegisterToolResponse.Status.Value("OK"):
        return None
    raise iterm2.rpc.RPCException(result.register_tool_response)



"""
The iTerm2 module provides a Python interface for controlling iTerm2.
"""
from iterm2.alert import Alert, TextInputAlert

from iterm2.app import (
    async_get_app, App, async_invoke_function,
    async_get_variable)

from iterm2.arrangement import SavedArrangementException, Arrangement

from iterm2.binding import PasteConfiguration, MoveSelectionUnit, SnippetIdentifier, BindingAction, KeyBinding, async_get_global_key_bindings, async_set_global_key_bindings, decode_key_binding

from iterm2.broadcast import BroadcastDomain, async_set_broadcast_domains

from iterm2.color import Color, ColorSpace

from iterm2.colorpresets import (
    ColorPreset, ListPresetsException, GetPresetException)

from iterm2.connection import Connection, run_until_complete, run_forever

from iterm2.customcontrol import CustomControlSequenceMonitor

from iterm2.focus import (
    FocusMonitor, FocusUpdateApplicationActive, FocusUpdateWindowChanged,
    FocusUpdateSelectedTabChanged, FocusUpdateActiveSessionChanged,
    FocusUpdate)

from iterm2.lifecycle import (
    EachSessionOnceMonitor, SessionTerminationMonitor, LayoutChangeMonitor,
    NewSessionMonitor)

from iterm2.mainmenu import MenuItemState, MainMenu, MenuItemException, MenuItemIdentifier

from iterm2.keyboard import (
    Modifier, Keycode, Keystroke, KeystrokePattern, KeystrokeMonitor,
    KeystrokeFilter)

from iterm2.preferences import PreferenceKey, async_get_preference, async_set_preference

from iterm2.profile import (
    Profile, PartialProfile, BadGUIDException, LocalWriteOnlyProfile,
    BackgroundImageMode, CursorType, ThinStrokes, UnicodeNormalization,
    CharacterEncoding, OptionKeySends, InitialWorkingDirectory, IconMode,
    TitleComponents, WriteOnlyProfile)

from iterm2.prompt import (
    Prompt, PromptMonitor, PromptState, async_get_last_prompt,
    async_list_prompts, async_get_prompt_by_id)

from iterm2.registration import RPC, ContextMenuProviderRPC, TitleProviderRPC, StatusBarRPC, Reference

from iterm2.screen import ScreenStreamer, LineContents, ScreenContents

from iterm2.selection import SelectionMode, SubSelection, Selection

from iterm2.session import (
    SplitPaneException, Splitter, Session, InvalidSessionId)

from iterm2.statusbar import (
    StatusBarComponent, CheckboxKnob, StringKnob, PositiveFloatingPointKnob,
    ColorKnob)

from iterm2.transaction import Transaction

from iterm2.tab import Tab, NavigationDirection

from iterm2.tmux import (
    TmuxException, TmuxConnection, async_get_tmux_connections,
    async_get_tmux_connection_by_connection_id)

from iterm2.tool import async_register_web_view_tool

from iterm2.triggers import decode_trigger, Trigger, AlertTrigger, AnnotateTrigger, BellTrigger, BounceTrigger, CaptureTrigger, CoprocessTrigger, HighlightLineTrigger, HighlightTrigger, HyperlinkTrigger, InjectTrigger, MarkTrigger, MuteCoprocessTrigger, PasswordTrigger, RPCTrigger, RunCommandTrigger, SendTextTrigger, SetDirectoryTrigger, SetHostnameTrigger, SetTitleTrigger, SetUserVariableTrigger, ShellPromptTrigger, StopTrigger, UserNotificationTrigger

from iterm2.util import (
    frame_str, size_str, Size, Point, Frame, CoordRange, Range,
    WindowedCoordRange, async_wait_forever)

from iterm2.window import (
    CreateTabException, CreateWindowException, SetPropertyException,
    GetPropertyException, Window)

from iterm2._version import __version__

from iterm2.rpc import RPCException

from iterm2.variables import VariableMonitor, VariableScopes



"""Enables showing modal alerts."""
import json
import typing

import iterm2.connection


class Alert:
    """A modal alert.

    :param title: The title, shown in bold at the top.
    :param subtitle: The informative text, which may be more than one line
        long.
    :param window_id: The window to attach the alert to. If None, it will be
        application modal.

    .. seealso:: Example ":ref:`oneshot_example`"
    """
    def __init__(
            self,
            title: str,
            subtitle: str,
            window_id: typing.Optional[str] = None):
        self.__title = title
        self.__subtitle = subtitle
        self.__buttons: typing.List[str] = []
        self.__window_id = window_id

    @property
    def title(self) -> str:
        """Returns the title string"""
        return self.__title

    @property
    def subtitle(self) -> str:
        """Returns the subtitle string"""
        return self.__subtitle

    @property
    def window_id(self) -> typing.Optional[str]:
        """Returns the window ID"""
        return self.__window_id

    def add_button(self, title: str):
        """Adds a button to the end of the list of buttons."""
        self.__buttons.append(title)

    async def async_run(self, connection: iterm2.connection.Connection) -> int:
        """Shows the modal alert.

        :param connection: The connection to use.
        :returns: The index of the selected button. If no buttons were defined
            then a single button, "OK", is automatically added.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.
        """
        title = json.dumps(self.title)
        subtitle = json.dumps(self.subtitle)
        buttons = json.dumps(self.__buttons)

        return await iterm2.async_invoke_function(
            connection,
            (f'iterm2.alert(title: {title}, ' +
             f'subtitle: {subtitle}, ' +
             f'buttons: {buttons}, ' +
             f'window_id: {json.dumps(self.window_id)})'))


class TextInputAlert:
    """A modal alert with a text input accessory.

    :param title: The title, shown in bold at the top.
    :param subtitle: The informative text, which may be more than one line
        long.
    :param placeholder: Grayed-out text to show in the text field when it is
        empty.
    :param default_value: Default text to place in the text field.
    :param window_id: Window ID to attach to, or None to make app-modal.
    """
    # pylint: disable=too-many-arguments
    def __init__(
            self,
            title: str,
            subtitle: str,
            placeholder: str,
            default_value: str,
            window_id: typing.Optional[str] = None):
        self.__title = title
        self.__subtitle = subtitle
        self.__placeholder = placeholder
        self.__default_value = default_value
        self.__window_id = window_id
    # pylint: enable=too-many-arguments

    @property
    def title(self) -> str:
        """Returns the title string"""
        return self.__title

    @property
    def subtitle(self) -> str:
        """Returns the subtitle string"""
        return self.__subtitle

    @property
    def placeholder(self) -> str:
        """Returns the placeholder string"""
        return self.__placeholder

    @property
    def default_value(self) -> str:
        """Returns the default value."""
        return self.__default_value

    # pylint: disable=invalid-name
    @property
    def defaultValue(self) -> str:
        """Deprecated in favor of default_valuedefault_value"""
        return self.__default_value
    # pylint: enable=invalid-name

    @property
    def window_id(self) -> typing.Optional[str]:
        """Returns the window ID"""
        return self.__window_id

    async def async_run(
            self,
            connection: iterm2.connection.Connection) -> typing.Optional[str]:
        """Shows the modal alert.

        :param connection: The connection to use.
        :returns: The string entered, or None if the alert was canceled.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.
        """
        title = json.dumps(self.title)
        subtitle = json.dumps(self.subtitle)
        placeholder = json.dumps(self.placeholder)
        default_value = json.dumps(self.default_value)

        return await iterm2.async_invoke_function(
            connection,
            (f'iterm2.get_string(title: {title}, ' +
             f'subtitle: {subtitle}, ' +
             f'placeholder: {placeholder}, ' +
             f'defaultValue: {default_value}, ' +
             f'window_id: {json.dumps(self.window_id)})'))



import AppKit
import Foundation
import __main__
import inspect
import os
import pathlib
import sys
import typing

class AuthenticationException(Exception):
    pass

def run_applescript(script):
    s = Foundation.NSAppleScript.alloc().initWithSource_(script)
    return s.executeAndReturnError_(None)

def get_string(result):
    return result[0].stringValue()

def get_error(result):
    return result[1]["NSAppleScriptErrorNumber"]

def has_error(result):
    return result[0] is None

def get_error_reason(result):
    return result[1]["NSAppleScriptErrorBriefMessage"]

def get_script_name():
    name = None
    if hasattr(__main__, "__file__"):
        fileName = __main__.__file__
        if os.path.basename(fileName) == "__main__.py":
            fileName = os.path.dirname(__main__.__file__)
        name = pathlib.Path(os.path.basename(fileName))
    elif len(sys.argv) > 0:
        name = os.path.basename(sys.argv[0])
    if not name:
        name = "Unknown"
    return name

def request_cookie_and_key(
        launch_if_needed: bool, myname: typing.Optional[str]):
    script = """
    set appName to "iTerm2"

    if application appName is running then
        return "yes"
    else
        return "no"
    end if
    """
    if not launch_if_needed:
        is_running = get_string(run_applescript(script))
        if is_running == "no":
            raise AuthenticationException("iTerm2 not running")
    justName = myname
    if justName is None:
        justName = get_script_name()
    s = Foundation.NSAppleScript.alloc().initWithSource_(
            'tell application "iTerm2" to request cookie and key ' +
            f'for app named "{justName}"')
    result = s.executeAndReturnError_(None)
    if has_error(result):
        if get_error(result) == -2740:
            raise AuthenticationException("iTerm2 version too old")

        reason = get_error_reason(result)
        raise AuthenticationException(reason)
    return get_string(result)

class LSBackgroundContextManager():
    def __init__(self):
        self.__value = AppKit.NSBundle.mainBundle().infoDictionary().get(
                "LSBackgroundOnly")

    def __enter__(self):
        info = AppKit.NSBundle.mainBundle().infoDictionary()
        info["LSBackgroundOnly"] = "1"

    def __exit__(self, exc_type, exc_value, exc_traceback):
        info = AppKit.NSBundle.mainBundle().infoDictionary()
        if self.__value:
            info["LSBackgroundOnly"] = __value
        else:
            del info["LSBackgroundOnly"]

def applescript_auth_disabled():
    filename = os.path.expanduser("~/Library/Application Support/iTerm2/disable-automation-auth")
    try:
        magic = "61DF88DC-3423-4823-B725-22570E01C027"
        expected = filename.encode("utf-8").hex() + " " + magic
        stat = os.stat(filename, follow_symlinks=False)
        if stat.st_uid != 0:
            return False
        if stat.st_size != len(expected):
            return False
        with open(filename, "r") as f:
            return f.read() == expected
    except:
        return False

def authenticate(
        launch_if_needed: bool = False,
        myname: typing.Optional[str] = None) -> bool:
    """Attempts to authenticate before connecting to iTerm2 API.

    :param launch_if_needed: If iTerm2 is not running, try to launch it first.
    :param myname: Name of this script to show in the console.

    :returns: True if a cookie and key were fetched. False if an existing one
        was kept.

    Raises an :class:`~iterm2.AuthenticationException` exception if
    authentication fails.
    """
    if applescript_auth_disabled():
        return True
    if os.environ.get("ITERM2_COOKIE"):
        return False
    with LSBackgroundContextManager() as _:
        cookie_and_key = request_cookie_and_key(launch_if_needed, myname)
        cookie, key = cookie_and_key.split(" ")
        os.environ["ITERM2_COOKIE"] = cookie
        os.environ["ITERM2_KEY"] = key
        return True



"""Provides a class that represents an iTerm2 tab."""
import abc
import enum
import json
import typing

import iterm2.api_pb2
import iterm2.capabilities
import iterm2.rpc
import iterm2.session
import iterm2.util


class NavigationDirection(enum.Enum):
    """Cardinal directions."""
    LEFT = "left"
    RIGHT = "right"
    ABOVE = "above"
    BELOW = "below"


class Tab:
    """Represents a tab.

    Don't create this yourself. Instead, use :class:`~iterm2.App`."""

    # pylint: disable=too-few-public-methods
    class Delegate:
        """Delegate for Tab."""
        @abc.abstractmethod
        def tab_delegate_get_window(
                self, tab: 'Tab') -> typing.Optional['iterm2.window.Window']:
            """Returns the Window for a Tab."""

        @abc.abstractmethod
        async def tab_delegate_get_window_by_id(
                self,
                window_id: str) -> typing.Optional['iterm2.window.Window']:
            """Returns the Window with the given ID."""
    # pylint: enable=too-few-public-methods

    delegate: typing.Optional[Delegate] = None

    # pylint: disable=too-many-arguments
    def __init__(
            self,
            connection,
            tab_id,
            root,
            tmux_window_id=None,
            tmux_connection_id=None,
            minimized_sessions=[]):
        self.connection = connection
        self.__tab_id = tab_id
        self.__root = root
        self.__active_session_id = None
        self.__tmux_window_id = tmux_window_id
        self.__tmux_connection_id = tmux_connection_id
        self.__minimized_sessions = list(minimized_sessions)
    # pylint: enable=too-many-arguments

    def __repr__(self):
        return "<Tab id=%s sessions=%s>" % (self.__tab_id, self.sessions)

    @property
    def active_session_id(self):
        return self.__active_session_id

    @active_session_id.setter
    def active_session_id(self, active_session_id):
        self.__active_session_id = active_session_id

    def update_from(self, other):
        """Copies state from another tab into this one."""
        self.__root = other.root
        self.__minimized_sessions = list(other.minimized_sessions)

    def update_session(self, session):
        """Replaces references to a session."""
        if self.__root.update_session(session):
            return
        indexes = [idx for idx, candidate in enumerate(self.__minimized_sessions) if candidate.session_id == session.session_id]
        if len(indexes) == 0:
            return
        i = indexes[0]
        self.__minimized_sessions[i] = session

    @property
    def window(self) -> typing.Optional['iterm2.window.Window']:
        """Returns the window this tab belongs to."""
        assert self.__class__.delegate
        return self.__class__.delegate.tab_delegate_get_window(self)

    @property
    def tmux_connection_id(self):
        """Returns the connection ID."""
        return self.__tmux_connection_id

    @property
    def tab_id(self) -> str:
        """
        Each tab has a globally unique identifier.

        :returns: The tab's identifier, a string.
        """
        return self.__tab_id

    @property
    def sessions(self) -> typing.List['iterm2.session.Session']:
        """
        A tab contains a list of sessions, which are its split panes.

        This excludes minimized sessions. Use `all_sessions` to get both visible
        and minimized sessions in this tab.

        :returns: The sessions belonging to this tab, in no particular order.
        """
        return self.__root.sessions

    @property
    def all_sessions(self) -> typing.List['iterm2.session.Session']:
        """
        Returns both visible and minimized sessions in this tab.

        A session would be minimized if another session in the tab is maximized.

        :returns: All sessions in this tab, including minimized sessions, in no
             particular order.
        """
        return self.sessions + self.minimized_sessions

    @property
    def root(self) -> iterm2.session.Splitter:
        """
        A tab's sessions are stored in a tree. This returns the root of that
        tree.

        An interior node of the tree is a Splitter. That corresponds to a
        collection of adjacent sessions with split pane dividers that are all
        either vertical or horizontal.

        Leaf nodes are Sessions.

        :returns: The root of the session tree.
        """
        return self.__root

    @property
    def current_session(self) -> typing.Union[None, iterm2.session.Session]:
        """
        :returns: The active session in this tab or `None` if it could not be
            determined.
        """
        for session in self.sessions:
            if session.session_id == self.active_session_id:
                return session
        return None

    @property
    def minimized_sessions(self) -> typing.List[iterm2.session.Session]:
        """
        :returns: Minimized sessions in this tab. Empty array if none.
        """
        return list(self.__minimized_sessions)

    def pretty_str(self, indent: str = "") -> str:
        """
        :returns: A human readable description of the tab and its sessions.
        """
        session = indent + "Tab id=%s\n" % self.tab_id
        session += self.__root.pretty_str(indent=indent + "  ")
        return session

    async def async_select(self, order_window_front: bool = True) -> None:
        """Deprecated in favor of `async_activate`."""
        await self.async_activate(order_window_front)

    async def async_activate(self, order_window_front: bool = True) -> None:
        """
        Selects this tab.

        :param order_window_front: Whether the window this session is in should
            be brought to the front and given keyboard focus.

        .. seealso:: Example ":ref:`function_key_tabs_example`"
        """
        await iterm2.rpc.async_activate(
            self.connection,
            False,
            True,
            order_window_front,
            tab_id=self.__tab_id)

    async def async_select_pane_in_direction(
            self, direction: NavigationDirection) -> typing.Optional[str]:
        """
        Activates a split pane adjacent to the currently selected pane.
        Requires iTerm2 version 3.3.2.

        :param direction: Specifies the direction to move. For example, LEFT
            will cause the pane to the left of the currently active one.
        :returns: The ID of the newly selected session ID, or None if there was
            no session in that direction.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.
        """
        if not iterm2.capabilities.supports_select_pane_in_direction(
                self.connection):
            raise iterm2.capabilities.AppVersionTooOld()

        invocation = iterm2.util.invocation_string(
            "iterm2.select_pane_in_direction",
            {"direction": direction.value})
        return await iterm2.rpc.async_invoke_method(
            self.connection, self.tab_id, invocation, -1)

    async def async_update_layout(self) -> None:
        """Adjusts the layout of the sessions in this tab.

        Change the `Session.preferred_size` of any sessions you wish to adjust
        before calling this.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.
        """
        response = await iterm2.rpc.async_set_tab_layout(
            self.connection, self.tab_id, self.__root.to_protobuf())
        status = response.set_tab_layout_response.status
        # pylint: disable=no-member
        if status == iterm2.api_pb2.SetTabLayoutResponse.Status.Value("OK"):
            return response.set_tab_layout_response
        raise iterm2.rpc.RPCException(
            iterm2.api_pb2.SetTabLayoutResponse.Status.Name(status))

    @property
    def tmux_window_id(self) -> typing.Union[None, str]:
        """Returns this tab's tmux window id or None.

        :returns: A tmux window id or `None` if this is not a tmux integration
            window.
        """
        return self.__tmux_window_id

    async def async_set_variable(self, name: str, value: typing.Any) -> None:
        """
        Sets a user-defined variable in the tab.

        See the Scripting Fundamentals documentation for more information on
        user-defined variables.

        :param name: The variable's name. Must begin with `user.`.
        :param value: The new value to assign.

        :throws: :class:`RPCException` if something goes wrong.
        """
        result = await iterm2.rpc.async_variable(
            self.connection,
            sets=[(name, json.dumps(value))],
            tab_id=self.__tab_id)
        status = result.variable_response.status
        # pylint: disable=no-member
        if status != iterm2.api_pb2.VariableResponse.Status.Value("OK"):
            raise iterm2.rpc.RPCException(
                iterm2.api_pb2.VariableResponse.Status.Name(status))

    async def async_get_variable(self, name: str) -> typing.Any:
        """
        Fetches a tab variable.

        See Badges documentation for more information on variables.

        :param name: The variable's name.

        :returns: The variable's value or `None` if it is undefined.

        :throws: :class:`RPCException` if something goes wrong.

        .. seealso:: Example ":ref:`sorttabs_example`"
        """
        result = await iterm2.rpc.async_variable(
            self.connection, gets=[name], tab_id=self.__tab_id)
        status = result.variable_response.status
        # pylint: disable=no-member
        if status != iterm2.api_pb2.VariableResponse.Status.Value("OK"):
            raise iterm2.rpc.RPCException(
                iterm2.api_pb2.VariableResponse.Status.Name(status))
        return json.loads(result.variable_response.values[0])

    async def async_close(self, force: bool = False) -> None:
        """
        Closes the tab.

        :param force: If True, the user will not be prompted for a
            confirmation.

        :throws: :class:`RPCException` if something goes wrong.

        .. seealso:: Example ":ref:`close_to_the_right_example`"
        """
        result = await iterm2.rpc.async_close(
            self.connection, tabs=[self.__tab_id], force=force)
        status = result.close_response.statuses[0]
        # pylint: disable=no-member
        if status != iterm2.api_pb2.CloseResponse.Status.Value("OK"):
            raise iterm2.rpc.RPCException(
                iterm2.api_pb2.CloseResponse.Status.Name(status))

    async def async_set_title(self, title: str):
        """Changes the tab's title.

        This is equivalent to editing the tab's title with the menu item Edit
        Tab Title. The title is an interpolated string.

        :param title: The new title. Set it to an empty string to use the
            default value (the current session's title).

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.
        """
        invocation = iterm2.util.invocation_string(
            "iterm2.set_title",
            {"title": title})
        await iterm2.rpc.async_invoke_method(
            self.connection, self.tab_id, invocation, -1)

    async def async_invoke_function(
            self, invocation: str, timeout: float = -1):
        """
        Invoke an RPC. Could be a registered function by this or another script
        of a built-in function.

        This invokes the RPC in the context of this tab. Note that most
        user-defined RPCs expect to be invoked in the context of a session.
        Default variables will be pulled from that scope. If you call a
        function from the wrong context it may fail because its defaults will
        not be set properly.

        :param invocation: A function invocation string.
        :param timeout: Max number of secondsto wait. Negative values mean to
            use the system default timeout.

        :returns: The result of the invocation if successful.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.
        """
        response = await iterm2.rpc.async_invoke_function(
            self.connection,
            invocation,
            tab_id=self.tab_id,
            timeout=timeout)
        which = response.invoke_function_response.WhichOneof('disposition')
        if which == 'error':
            # pylint: disable=no-member
            if (response.invoke_function_response.error.status ==
                    iterm2.api_pb2.InvokeFunctionResponse.Status.
                    Value("TIMEOUT")):
                raise iterm2.rpc.RPCException("Timeout")
            raise iterm2.rpc.RPCException("{}: {}".format(
                iterm2.api_pb2.InvokeFunctionResponse.Status.Name(
                    response.invoke_function_response.error.status),
                response.invoke_function_response.error.error_reason))
        return json.loads(
            response.invoke_function_response.success.json_result)

    async def async_move_to_window(self) -> 'iterm2.window.Window':
        """
        Moves this tab to its own window, provided there are multiple tabs in
        the window it belongs to.

        :returns: The new window ID.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.
        """
        window_id = await self.async_invoke_function(
            "iterm2.move_tab_to_window()")
        assert self.__class__.delegate
        window = await self.__class__.delegate.tab_delegate_get_window_by_id(
            window_id)
        if not window:
            raise iterm2.rpc.RPCException(
                "No such window {}".format(window_id))
        return window



"""Provides handy functions."""
import asyncio
import json
import typing

import iterm2.api_pb2

# pylint: disable=invalid-name
class Size:
    """Describes a 2D size.

    Can be used where api_pb2.Size is accepted."""

    def __init__(self, width: int, height: int):
        """Constructs a new size.

        :param width: A nonnegative number giving the width.
        :param height: A nonnegative number giving the height.
        """
        self.__width = width
        self.__height = height

    @property
    def width(self) -> int:
        """Gets the width"""
        return self.__width

    @width.setter
    def width(self, value: int):
        """Sets the width"""
        self.__width = value

    @property
    def height(self) -> int:
        """Gets the height"""
        return self.__height

    @height.setter
    def height(self, value: int):
        """Sets the height"""
        self.__height = value

    @property
    def dict(self):
        """
        Returns a dict representation of the size.
        """
        return {"width": self.width, "height": self.height}

    def load_from_dict(self, dictionary):
        """
        Initializes the size from a dict representation.
        """
        self.width = dictionary["width"]
        self.height = dictionary["height"]

    @property
    def json(self) -> str:
        """
        Gives a JSON representation of the size.
        """
        return json.dumps(self.dict)

    @property
    def proto(self):
        """Returns the protobuf Size representation."""
        size = iterm2.api_pb2.Size()
        size.width = self.width
        size.height = self.height
        return size


class Point:
    """Describes a 2D coordinate.

    Can be used where api_pb2.Point is accepted."""

    def __init__(self, x: int, y: int):
        """Constructs a new point.

        :param x: A number giving the X coordinate.
        :param y: A number giving the Y coordinate.
        """
        self.__x = x
        self.__y = y

    def __repr__(self):
        return "({}, {})".format(self.x, self.y)

    @staticmethod
    def from_coord_proto(proto):
        """Creates a new Point from a Coord protobuf"""
        return Point(proto.x, proto.y)

    @property
    def x(self) -> int:
        """Returns the x coordinate"""
        return self.__x

    @x.setter
    def x(self, value: int):
        """Sets the x coordinate"""
        self.__x = value

    @property
    def y(self) -> int:
        """Returns the y coordinate"""
        return self.__y

    @y.setter
    def y(self, value: int):
        """Sets the y coordinate"""
        self.__y = value

    @property
    def dict(self):
        """Returns a dict representation of the point."""
        return {"x": self.x, "y": self.y}

    def load_from_dict(self, dict):  # pylint: disable=redefined-builtin
        """Initializes the point from a dict representation."""
        self.x = dict["x"]
        self.y = dict["y"]

    @property
    def json(self) -> str:
        """Returns a JSON representation of the point."""
        return json.dumps(self.dict)

    @property
    def proto(self):
        """Returns the protobuf Coord representation."""
        coord = iterm2.api_pb2.Coord()
        coord.x = self.x
        coord.y = self.y
        return coord

    def __eq__(self, other):
        if isinstance(other, Point):
            return self.x == other.x and self.y == other.y
        return NotImplemented

    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.items())))


class Frame:
    """Describes a bounding rectangle. 0,0 is the bottom left coordinate."""
    def __init__(self, origin: Point = Point(0, 0), size: Size = Size(0, 0)):
        """Constructs a new frame."""
        self.__origin = origin
        self.__size = size

    def __repr__(self):
        return "<Frame origin={} size={}>".format(self.origin, self.size)

    @property
    def origin(self) -> Point:
        """The top-left coordinate.

        :returns: A :class:`Point`.
        """
        return self.__origin

    @origin.setter
    def origin(self, value: Point):
        """Sets the top-left coordinate.

        :param value: A :class:`Point`.
        """
        self.__origin = value

    @property
    def size(self) -> Size:
        """The size.

        :returns: A :class:`Size`.
        """
        return self.__size

    @size.setter
    def size(self, value: Size):
        """Sets the size.

        :param value: A :class:`Size`.
        """
        self.__size = value

    def load_from_dict(self, dict):  # pylint: disable=redefined-builtin
        """Sets the frame's values from a dict representation."""
        self.origin.load_from_dict(dict["origin"])
        self.size.load_from_dict(dict["size"])

    @property
    def dict(self):
        """Returns a dict representation of the frame."""
        return {"origin": self.origin.dict, "size": self.size.dict}

    @property
    def json(self) -> str:
        """Returns a JSON representation of the frame."""
        return json.dumps(self.dict)


def frame_str(frame: Frame) -> str:
    """Formats a :class:`Frame` as a human-readable string.

    :param frame: The frame to convert.

    :returns: A human-readable string."""
    if frame is None:
        return "[Undefined]"

    return "[(%s, %s) %s]" % (
        frame.origin.x,
        frame.origin.y,
        size_str(frame.size))


def size_str(size: Size) -> str:
    """Formats a :class:`Size` as a human-readable string.

    :param size: The size to convert.

    :returns: A human-readable string."""
    if size is None:
        return "[Undefined]"
    return "(%s x %s)" % (
        size.width,
        size.height)


def point_str(point: Point) -> str:
    """Formats a :class:`Point` as a human-readable string.

    :param point: The point to convert.

    :returns: A human-readable string."""
    if point is None:
        return "[Undefined]"
    return "(%s, %s)" % (point.x, point.y)


def distance(point_a, point_b, grid_width):
    """
    Returns the distance from Point point_a to Point point_b, as measured by
    the number of cells in left-to-right, top-to-bottom order.
    """
    a_pos = point_a.y
    a_pos *= grid_width
    a_pos += point_a.x

    b_pos = point_b.y
    b_pos *= grid_width
    b_pos += point_b.x

    return abs(a_pos - b_pos)


class CoordRange:
    """Describes a range of contiguous cells.

    :param start: The start point.
    :param end: The first point after the start point not in the range."""
    def __init__(self, start: Point, end: Point):
        self.__start = start
        self.__end = end

    def __repr__(self):
        return "CoordRange({} to {})".format(self.start, self.end)

    @staticmethod
    def from_proto(proto):
        """Creates a new CoordRange from a protobuf"""
        return CoordRange(
            Point.from_coord_proto(proto.start),
            Point.from_coord_proto(proto.end))

    @property
    def start(self) -> Point:
        """:returns: The start :class:`Point`."""
        return self.__start

    @property
    def end(self) -> Point:
        """
        :returns: The first :class:`Point` after `self.start` not in the
            range."""
        return self.__end

    @property
    def proto(self):
        """Returns the protobuf CoordRange representation."""
        # pylint: disable=no-member
        coord = iterm2.api_pb2.CoordRange()
        coord.start.CopyFrom(self.start.proto)
        coord.end.CopyFrom(self.end.proto)
        return coord

    def length(self, width) -> int:
        """Returns the number of cells from start to end."""
        return distance(self.start, self.end, width)


class Range:
    """Describes a range of integers.

    :param location: The first value in the range.
    :param length: The number of values in the range."""
    def __init__(self, location: int, length: int):
        self.__location = location
        self.__length = length

    def __repr__(self):
        return "[{}, {})".format(self.location, self.location + self.length)

    @property
    def location(self) -> int:
        """:returns: The first location of the range."""
        return self.__location

    @property
    def length(self) -> int:
        """:returns: The length of the range."""
        return self.__length

    @property
    def max(self) -> int:
        """Returns the first value after the range."""
        return self.location + self.length

    @property
    def proto(self):
        """Returns the protobuf Range representation."""
        protobuf = iterm2.api_pb2.Range()
        protobuf.location = self.location
        protobuf.length = self.length
        return protobuf

    @property
    def toSet(self):
        """Deprecated. Use to_set instead."""
        return self.to_set

    @property
    def to_set(self):
        """Returns a `set` containing all values in the range."""
        return set(range(self.location, self.location + self.length))


class WindowedCoordRange:
    """
    Describes a range of coordinates, optionally constrained to a continugous
    range of columns.

    :param coordRange: The range of cells.
    :param columnRange: The range of columns to intersect with `coordRange` to
        get the described region, or `None` if unwindowed.
    """
    def __init__(
            self,
            coordRange: CoordRange,
            columnRange: typing.Optional[Range] = None):
        self.___coord_range = coordRange
        if columnRange:
            self.__column_range = columnRange
        else:
            self.__column_range = Range(0, 0)

    def __repr__(self):
        return "WindowedCoordRange(coordRange={} cols={})".format(
            self.coordRange, self.columnRange)

    @property
    def coordRange(self) -> CoordRange:
        """:returns: The range of coordinates."""
        return self.___coord_range

    @property
    def columnRange(self) -> Range:
        """
        :returns: The range of columns, or an empty range if unconstrained.
        """
        return self.__column_range

    @property
    def proto(self):
        """Returns the protobuf WindowedCoordRange representation."""
        # pylint: disable=no-member
        windowed_coord_range = iterm2.api_pb2.WindowedCoordRange()
        windowed_coord_range.coord_range.CopyFrom(self.coordRange.proto)
        windowed_coord_range.columns.CopyFrom(self.columnRange.proto)
        return windowed_coord_range

    @property
    def start(self) -> Point:
        """The first point in the region."""
        x, y = self.coordRange.start.x, self.coordRange.start.y
        if self.columnRange.length:
            x = min(max(x, self.columnRange.location),
                    self.columnRange.location + self.columnRange.length)
        return Point(x, y)

    @property
    def end(self) -> Point:
        """The first point not in the region."""
        x, y = self.coordRange.end.x, self.coordRange.end.y
        if self.hasWindow:
            x = min(self.coordRange.end.x, self.right + 1)
        return Point(x, y)

    @property
    def right(self) -> int:
        """The rightmost column allowed, or 0 if unwindowed."""
        return self.__column_range.location + self.__column_range.length

    @property
    def left(self) -> int:
        """The leftmost column allowed."""
        if self.hasWindow:
            return self.__column_range.location
        return 0

    @property
    def hasWindow(self) -> bool:
        """Deprecated. Use has_window instead."""
        return self.has_window

    @property
    def has_window(self) -> bool:
        """Are the columns constrained?"""
        return self.__column_range.length > 0



async def async_wait_forever():
    """A convenience function that never returns."""
    await asyncio.wait([asyncio.Future()])


def invocation_string(
        method_name: str, argdict: typing.Dict[str, typing.Any]) -> str:
    """Gives the invocation string for a method call with given arguments."""
    parts = []
    for name, value in argdict.items():
        parts.append(f"{name}: {json.dumps(value)}")
    return method_name + "(" + ", ".join(parts) + ")"



"""Provides access to saved arrangements."""

import iterm2.capabilities
import iterm2.connection
import iterm2.rpc
import iterm2.api_pb2
import typing


class SavedArrangementException(Exception):
    """A problem was encountered while saving or restoring an arrangement."""


class Arrangement:
    """Provides access to saved arrangements."""
    @staticmethod
    async def async_save(connection: iterm2.connection.Connection, name: str):
        """Save all windows as a new arrangement.

        Replaces the arrangement with the given name if it already exists.

        :param connection: The name of the arrangement.
        :param name: The name of the arrangement.

        :throws: SavedArrangementException
        """
        result = await iterm2.rpc.async_save_arrangement(connection, name)
        status = result.saved_arrangement_response.status
        # pylint: disable=no-member
        if status != iterm2.api_pb2.CreateTabResponse.Status.Value("OK"):
            raise SavedArrangementException(
                iterm2.api_pb2.SavedArrangementResponse.Status.Name(
                    result.saved_arrangement_response.status))

    @staticmethod
    async def async_restore(
        connection: iterm2.connection.Connection,
        name: str,
        window_id: typing.Optional[str] = None):
        """Restore a saved window arrangement.

        :param connection: The name of the arrangement.
        :param name: The name of the arrangement to restore.
        :param window_id: The window to restore in to as tabs, or None to
             restore as new windows.

        :throws: SavedArrangementException
        """
        result = await iterm2.rpc.async_restore_arrangement(connection, name, window_id)
        status = result.saved_arrangement_response.status
        # pylint: disable=no-member
        if status != iterm2.api_pb2.CreateTabResponse.Status.Value("OK"):
            raise SavedArrangementException(
                iterm2.api_pb2.SavedArrangementResponse.Status.Name(
                    result.saved_arrangement_response.status))

    @staticmethod
    async def async_list(
            connection: iterm2.connection.Connection) -> typing.List[str]:
        """Fetch a list of saved arrangements.

        NOTE: This requires iTerm2 version 3.4.0 or later.

        :param connection: The name of the arrangement.

        :returns: A list of strings giving saved arrangement names.

        :throws: SavedArrangementException
        """
        # Throw if iTerm2 is too old to support this feature.
        iterm2.capabilities.check_supports_list_saved_arrangements(connection)

        # Send the request
        result = await iterm2.rpc.async_list_arrangements(connection)
        status = result.saved_arrangement_response.status

        # pylint: disable=no-member
        if status != iterm2.api_pb2.CreateTabResponse.Status.Value("OK"):
            raise SavedArrangementException(
                iterm2.api_pb2.SavedArrangementResponse.Status.Name(
                    result.saved_arrangement_response.status))
        return result.saved_arrangement_response.names




"Abstractions for iTerm2 triggers."

import enum
import iterm2
import typing

def _hex(color):
    if not color:
        return ""
    return color.hex

def _futureproof(param, obj):
    """Sets the unparsed parameter in a trigger.

    Subclasses of Trigger will initialize __param with a computed param. This
    limits it to features that are parseable. If new values are added to a param
    in the future (e.g., color changes from {text,background} to
    {text,background,colorspace}) then those future features would be able to
    round trip through HighlightTrigger.deserialize().param. This function is a
    convenience to replace __param with the raw value so that as long as you
    don't modify a trigger it will produce the same param it was initialized
    with.
    """
    obj.param = param
    return obj

def decode_trigger(encoded: dict) -> 'Trigger':
    """Create a trigger.

    Use this to convert the dictionary representation of a trigger gotten from
    :class:`~iterm2.Profile` and friends into a :class:`~Trigger` object.

    :param encoded: The encoded trigger.

    :returns: A :class:`~Trigger`.
    """
    classes = {
        AlertTrigger._name(): AlertTrigger,
        AnnotateTrigger._name(): AnnotateTrigger,
        BellTrigger._name(): BellTrigger,
        BounceTrigger._name(): BounceTrigger,
        CaptureTrigger._name(): CaptureTrigger,
        CoprocessTrigger._name(): CoprocessTrigger,
        HighlightLineTrigger._name(): HighlightLineTrigger,
        HighlightTrigger._name(): HighlightTrigger,
        HyperlinkTrigger._name(): HyperlinkTrigger,
        InjectTrigger._name(): InjectTrigger,
        MarkTrigger._name(): MarkTrigger,
        MuteCoprocessTrigger._name(): MuteCoprocessTrigger,
        PasswordTrigger._name(): PasswordTrigger,
        RPCTrigger._name(): RPCTrigger,
        RunCommandTrigger._name(): RunCommandTrigger,
        SendTextTrigger._name(): SendTextTrigger,
        SetDirectoryTrigger._name(): SetDirectoryTrigger,
        SetHostnameTrigger._name(): SetHostnameTrigger,
        SetTitleTrigger._name(): SetTitleTrigger,
        SetUserVariableTrigger._name(): SetUserVariableTrigger,
        ShellPromptTrigger._name(): ShellPromptTrigger,
        StopTrigger._name(): StopTrigger,
        UserNotificationTrigger._name(): UserNotificationTrigger,
    }

    name = encoded["action"]
    regex = encoded["regex"]
    param = encoded.get("parameter", "")
    instant = encoded.get("partial", False)
    enabled = not encoded.get("disabled", False)

    if name not in classes:
        # Futureproof unrecognized trigger types. This allows a round-trip
        # through the Trigger representation.
        return Trigger(regex, param, instant, enabled)

    return classes[name].deserialize(regex, param, instant, enabled)

class Trigger:
    """Base class for triggers.

    Do not create this yourself. Instead, create a subclass directly.

    You may get an instance of this object for unrecognized trigger types that
    are defined in future versions of iTerm2.
    """
    def __init__(self, regex: str, param, instant: bool, enabled: bool):
        self.__regex = regex
        self.__param = param
        self.__instant = instant
        self.__enabled = enabled

    def __repr__(self):
        return f'<{self.__class__.__name__}: regex={self.__regex} instant={self.__instant} enabled={self.__enabled} param={self._param}>'

    def __eq__(self, other):
        return (
             self.__regex == other.__regex and
             self.__param == other.__param and
             self.__instant == other.__instant and
             self.__enabled == other.__enabled)

    @property
    def param(self):
        return self.__param

    @param.setter
    def param(self, value):
        self.__param = value

    @property
    def regex(self) -> str:
        return self.__regex

    @regex.setter
    def regex(self, value: str):
        self.__regex = value

    @property
    def instant(self) -> bool:
        return self.__instant

    @instant.setter
    def instant(self, value: bool):
        self.__instant = value

    @property
    def enabled(self) -> bool:
        return self.__enabled

    @enabled.setter
    def enabled(self, value: bool):
        self.__enabled = value

    @property
    def encode(self) -> dict:
        return { "regex":  self.regex,
                 "action":  self._name(),
                 "parameter": self.param,
                 "partial": self.instant,
                 "disabled": not self.enabled }

    def toJSON(self):
        return json.dumps(self.encode)

    @property
    def _param(self):
        return ""


class AlertTrigger(Trigger):
    def __init__(self, regex: str, message: str, instant: bool, enabled: bool):
        self.__message = message
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "AlertTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        return _futureproof(param, AlertTrigger(regex, param, instant, enabled))

    @property
    def message(self) -> str:
        return self.__message

    @message.setter
    def message(self, value: str):
        self.__message = value
        self.param = self._param

    @property
    def _param(self):
        return self.__message

class AnnotateTrigger(Trigger):
    def __init__(self, regex: str, annotation: str, instant: bool, enabled: bool):
        self.__annotation = annotation
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "AnnotateTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        return _futureproof(param, AnnotateTrigger(regex, param, instant, enabled))

    @property
    def annotation(self) -> str:
        return self.__annotation

    @annotation.setter
    def annotation(self, value: str):
        self.__annotation = value
        self.param = self._param

    @property
    def _param(self):
        return self.__annotation

class BellTrigger(Trigger):
    def __init__(self, regex: str, instant: bool, enabled: bool):
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "BellTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        return _futureproof(param, BellTrigger(regex, instant, enabled))

    @property
    def _param(self) -> None:
        return ""

class BounceTrigger(Trigger):
    class Action(enum.Enum):
        BOUNCE_UNTIL_ACTIVATED = 0
        BOUNCE_ONCE = 1

    def __init__(self, regex: str, action: 'iterm2.Trigger.BounceTrigger.Action', instant: bool, enabled: bool):
        self.__action = action
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "BounceTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        return _futureproof(param, BounceTrigger(regex, BounceTrigger.Action(param), instant, enabled))

    @property
    def action(self) -> 'iterm2.Trigger.BounceTrigger.Action':
        return self.__action

    @action.setter
    def action(self, value: 'iterm2.Trigger.BounceTrigger.Action'):
        self.__action = value
        self.param = self._param

    @property
    def _param(self):
        return self.__action.value

class RPCTrigger(Trigger):
    def __init__(self, regex: str, invocation: str, instant: bool, enabled: bool):
        self.__invocation = invocation
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "iTermRPCTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        return _futureproof(param, RPCTrigger(regex, param, instant, enabled))

    @property
    def invocation(self) -> str:
        return self.__invocation

    @invocation.setter
    def invocation(self, value: str):
        self.__invocation = value
        self.param = self._param

    @property
    def _param(self):
        return self.__invocation

class CaptureTrigger(Trigger):
    def __init__(self, regex: str, command: str, instant: bool, enabled: bool):
        self.__command = command
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "CaptureTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        return _futureproof(param, CaptureTrigger(regex, param, instant, enabled))

    @property
    def command(self) -> str:
        return self.__command

    @command.setter
    def command(self, value: str):
        self.__command = value
        self.param = self._param

    @property
    def _param(self):
        return self.__command

class InjectTrigger(Trigger):
    def __init__(self, regex: str, injection: str, instant: bool, enabled: bool):
        self.__injection = injection
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "iTermInjectTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        return _futureproof(param, InjectTrigger(regex, param, instant, enabled))

    @property
    def injection(self) -> str:
        return self.__injection

    @injection.setter
    def injection(self, value: str):
        self.__injection = value
        self.param = self._param

    @property
    def _param(self):
        return self.__injection

class HighlightLineTrigger(Trigger):
    def __init__(self, regex: str, text_color: typing.Optional[iterm2.Color], background_color: typing.Optional[iterm2.Color], instant: bool, enabled: bool):
        self.__text_color = text_color
        self.__background_color = background_color
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "iTermHighlightLineTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        parts = param[1:-1].split(",")
        text_color = iterm2.Color.from_trigger(parts[0])
        background_color = iterm2.Color.from_trigger(parts[1])
        return _futureproof(param, HighlightTrigger(regex, text_color, background_color, instant, enabled))

    @property
    def text_color(self) -> typing.Optional[iterm2.Color]:
        return self.__text_color

    @text_color.setter
    def text_color(self, value: typing.Optional[iterm2.Color]):
        self.__text_color = value
        self.param = self._param

    @property
    def background_color(self) -> typing.Optional[iterm2.Color]:
        return self.__background_color

    @background_color.setter
    def background_color(self, value: typing.Optional[iterm2.Color]):
        self.__background_color = value
        self.param = self._param

    @property
    def _param(self):
        return "{" + _hex(self.__text_color) + "," + _hex(self.__background_color) + "}"

class UserNotificationTrigger(Trigger):
    def __init__(self, regex: str, message: str, instant: bool, enabled: bool):
        self.__message = message
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "iTermUserNotificationTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        return _futureproof(param, UserNotificationTrigger(regex, param, instant, enabled))

    @property
    def message(self) -> str:
        return self.__message

    @message.setter
    def message(self, value: str):
        self.__message = value
        self.param = self._param

    @property
    def _param(self):
        return self.__message

class SetUserVariableTrigger(Trigger):
    def __init__(self, regex: str, name: str, json_value: str, instant: bool, enabled: bool):
        self.__name = name
        self.__json_value = json_value
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "iTermSetUserVariableTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        parts = param.split(chr(1))
        return _futureproof(param, SetUserVariableTrigger(regex, parts[0], parts[1], instant, enabled))

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, value: str):
        self.__name = value
        self.param = self._param

    @property
    def json_value(self) -> str:
        return self.__json_value

    @json_value.setter
    def json_value(self, value: str):
        self.__json_value = value
        self.param = self._param

    @property
    def _param(self):
        return self.__name + chr(1) + self.__json_value

class ShellPromptTrigger(Trigger):
    def __init__(self, regex: str, instant: bool, enabled: bool):
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "iTermShellPromptTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        return _futureproof(param, ShellPromptTrigger(regex, instant, enabled))

    @property
    def _param(self):
        return ""

class SetTitleTrigger(Trigger):
    def __init__(self, regex: str, title: str, instant: bool, enabled: bool):
        self.__title = title
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "iTermSetTitleTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        return _futureproof(param, SetTitleTrigger(regex, param, instant, enabled))

    @property
    def title(self) -> str:
        return self.__title

    @title.setter
    def title(self, value: str):
        self.__title = value
        self.param = self._param

    @property
    def _param(self):
        return self.__title

class SendTextTrigger(Trigger):
    def __init__(self, regex: str, text: str, instant: bool, enabled: bool):
        self.__text = text
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "SendTextTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        return _futureproof(param, SendTextTrigger(regex, param, instant, enabled))

    @property
    def text(self) -> str:
        return self.__text

    @text.setter
    def text(self, value: str):
        self.__text = value
        self.param = self._param

    @property
    def _param(self):
        return self.__text

class RunCommandTrigger(Trigger):
    def __init__(self, regex: str, command: str, instant: bool, enabled: bool):
        self.__command = command
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "ScriptTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        return _futureproof(param, RunCommandTrigger(regex, param, instant, enabled))

    @property
    def command(self) -> str:
        return self.__command

    @command.setter
    def command(self, value: str):
        self.__command = value
        self.param = self._param

    @property
    def _param(self):
        return self.__command

class CoprocessTrigger(Trigger):
    def __init__(self, regex: str, command: str, instant: bool, enabled: bool):
        self.__command = command
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "CoprocessTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        return _futureproof(param, CoprocessTrigger(regex, param, instant, enabled))

    @property
    def command(self) -> str:
        return self.__command

    @command.setter
    def command(self, value: str):
        self.__command = value
        self.param = self._param

    @property
    def _param(self):
        return self.__command

class MuteCoprocessTrigger(Trigger):
    def __init__(self, regex: str, command: str, instant: bool, enabled: bool):
        self.__command = command
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "MuteCoprocessTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        return _futureproof(param, MuteCoprocessTrigger(regex, param, instant, enabled))

    @property
    def command(self) -> str:
        return self.__command

    @command.setter
    def command(self, value: str):
        self.__command = value
        self.param = self._param

    @property
    def _param(self):
        return self.__command

class HighlightTrigger(Trigger):
    def __init__(self, regex: str, text_color: typing.Optional[iterm2.Color], background_color: typing.Optional[iterm2.Color], instant: bool, enabled: bool):
        self.__text_color = text_color
        self.__background_color = background_color
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "HighlightTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        parts = param[1:-1].split(",")
        text_color = iterm2.Color.from_trigger(parts[0])
        background_color = iterm2.Color.from_trigger(parts[1])
        return _futureproof(param, HighlightTrigger(regex, text_color, background_color, instant, enabled))

    @property
    def text_color(self) -> typing.Optional[iterm2.Color]:
        return self.__text_color

    @text_color.setter
    def text_color(self, value: typing.Optional[iterm2.Color]):
        self.__text_color = value
        self.param = self._param

    @property
    def background_color(self, value: typing.Optional[iterm2.Color]):
        return self.__background_color

    @background_color.setter
    def background_color(self, value: typing.Optional[iterm2.Color]):
        self.__background_color = value
        self.param = self._param

    @property
    def _param(self):
        return "{" + _hex(self.__text_color) + "," + _hex(self.__background_color) + "}"

class MarkTrigger(Trigger):
    def __init__(self, regex: str, stop_scrolling: bool, instant: bool, enabled: bool):
        self.__stop_scrolling = stop_scrolling
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "MarkTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        return _futureproof(param, MarkTrigger(regex, param == 1, instant, enabled))

    @property
    def stop_scrolling(self) -> bool:
        return self.__stop_scrolling

    @stop_scrolling.setter
    def stop_scrolling(self, value: bool):
        self.__stop_scrolling = value
        self.param = self._param

    @property
    def _param(self):
        return 1 if self.__stop_scrolling else 0

class PasswordTrigger(Trigger):
    SEPARATOR = "\u2002\u2014\u2002"

    def __init__(self, regex: str, account_name: str, user_name: typing.Optional[str], instant: bool, enabled: bool):
        self.__account_name = account_name
        self.__user_name = user_name
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "PasswordTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        if PasswordTrigger.SEPARATOR in param:
            parts = param.split(PasswordTrigger.SEPARATOR)
            account_name = parts[0]
            user_name = parts[1]
        else:
            account_name = param
            user_name = ""
        return _futureproof(param, PasswordTrigger(regex, account_name, user_name, instant, enabled))

    @property
    def account_name(self) -> str:
        return self.__account_name

    @account_name.setter
    def account_name(self, value: str):
        self.__account_name = value
        self.param = self._param

    @property
    def user_name(self) -> str:
        return self.__user_name

    @user_name.setter
    def user_name(self, value: str):
        self.__user_name = value
        self.param = self._param

    @property
    def _param(self):
        if len(self.__user_name) > 0:
            return self.__account_name + PasswordTrigger.SEPARATOR + self.__user_name
        return self.__account_name

class HyperlinkTrigger(Trigger):
    def __init__(self, regex: str, url: str, instant: bool, enabled: bool):
        self.__url = url
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "iTermHyperlinkTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        return _futureproof(param, HyperlinkTrigger(regex, param, instant, enabled))

    @property
    def url(self) -> str:
        return self.__url

    @url.setter
    def url(self, value: str):
        self.__url = value
        self.param = self._param

    @property
    def _param(self):
        return self.__url

class SetDirectoryTrigger(Trigger):
    def __init__(self, regex: str, directory: str, instant: bool, enabled: bool):
        self.__directory = directory
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "SetDirectoryTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        return _futureproof(param, SetDirectoryTrigger(regex, param, instant, enabled))

    @property
    def directory(self) -> str:
        return self.__directory

    @directory.setter
    def directory(self, value: str):
        self.__directory = value
        self.param = self._param

    @property
    def _param(self):
        return self.__directory

class SetHostnameTrigger(Trigger):
    def __init__(self, regex: str, hostname: str, instant: bool, enabled: bool):
        self.__hostname = hostname
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "SetHostnameTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        return _futureproof(param, SetHostnameTrigger(regex, param, instant, enabled))

    @property
    def hostname(self) -> str:
        return self.__hostname

    @hostname.setter
    def hostname(self, value: str):
        self.__hostname = value
        self.param = self._param

    @property
    def _param(self):
        return self.__hostname

class StopTrigger(Trigger):
    def __init__(self, regex: str, instant: bool, enabled: bool):
        super().__init__(regex, self._param, instant, enabled)

    @staticmethod
    def _name():
        return "StopTrigger"

    @staticmethod
    def deserialize(regex: str, param: str, instant: bool, enabled: bool):
        return _futureproof(param, StopTrigger(regex, instant, enabled))

    @property
    def _param(self):
        return ""




"""Provides classes for interacting with iTerm2 sessions."""
import abc
import json
import typing

import iterm2.api_pb2
import iterm2.capabilities
import iterm2.connection
import iterm2.keyboard
import iterm2.notifications
import iterm2.profile
import iterm2.rpc
import iterm2.screen
import iterm2.selection
import iterm2.util


# pylint: disable=too-many-lines
# pylint: disable=too-many-public-methods


class SplitPaneException(Exception):
    """Something went wrong when trying to split a pane."""


class Splitter:
    """
    A container of split pane sessions where the dividers are all aligned the
    same way.
    """

    def __init__(self, vertical: bool = False):
        """
        :param vertical: bool. If true, the divider is vertical, else
            horizontal.
        """
        self.__vertical = vertical
        # Elements are either Splitter or Session
        self.__children: typing.List[typing.Union['Splitter', 'Session']] = []

    @staticmethod
    def from_node(node, connection):
        """Creates a new Splitter from a node.

        :param node: :class:`iterm2.api_pb2.SplitTreeNode`
        :param connection: :class:`~iterm2.connection.Connection`

        :returns: A new Splitter.
        """
        splitter = Splitter(node.vertical)
        for link in node.links:
            if link.HasField("session"):
                session = Session(connection, link)
                splitter.add_child(session)
            else:
                subsplit = Splitter.from_node(link.node, connection)
                splitter.add_child(subsplit)
        return splitter

    @property
    def vertical(self) -> bool:
        """Are the dividers in this splitter vertical?"""
        return self.__vertical

    def add_child(self, child: typing.Union['Splitter', 'Session']):
        """
        Adds one or more new sessions to a splitter.

        child: A Session or a Splitter.
        """
        self.__children.append(child)

    @property
    def children(self) -> typing.List[typing.Union['Splitter', 'Session']]:
        """
        :returns: This splitter's children. A list of :class:`Session` or
            :class:`Splitter` objects.
        """
        return self.__children

    @property
    def sessions(self) -> typing.List['Session']:
        """
        :returns: All sessions in this splitter and all nested splitters. A
            list of :class:`Session` objects.
        """
        result = []
        for child in self.__children:
            if isinstance(child, Session):
                result.append(child)
            else:
                result.extend(child.sessions)
        return result

    def pretty_str(self, indent: str = "") -> str:
        """
        :returns: A string describing this splitter. Has newlines.
        """
        string_value = indent + "Splitter %s\n" % (
            "|" if self.vertical else "-")
        for child in self.__children:
            string_value += child.pretty_str("  " + indent)
        return string_value

    def update_session(self, session):
        """
        Finds a session with the same ID as session. If it exists, replace the
            reference with session.

        :returns: True if the update occurred.
        """
        i = 0
        for child in self.__children:
            if isinstance(
                    child, Session) and child.session_id == session.session_id:
                self.__children[i] = session
                return True
            if isinstance(child, Splitter):
                if child.update_session(session):
                    return True
            i += 1
        return False

    def to_protobuf(self):
        """Returns the protobuf representation."""
        node = iterm2.api_pb2.SplitTreeNode()
        node.vertical = self.vertical

        # pylint: disable=no-member
        def make_link(obj):
            link = iterm2.api_pb2.SplitTreeNode.SplitTreeLink()
            if isinstance(obj, Session):
                link.session.CopyFrom(obj.to_session_summary_protobuf())
            else:
                link.node.CopyFrom(obj.to_protobuf())
            return link

        links = list(map(make_link, self.children))
        node.links.extend(links)
        return node


class SessionLineInfo:
    """Describes a session's geometry."""
    def __init__(self, line_info):
        self.__line_info = line_info

    @property
    def mutable_area_height(self) -> int:
        """Returns the height of the mutable area of the session."""
        return self.__line_info[0]

    @property
    def scrollback_buffer_height(self) -> int:
        """Returns the height of the immutable area of the session."""
        return self.__line_info[1]

    @property
    def overflow(self) -> int:
        """
        Returns the number of lines lost to overflow. These lines were
        removed after scrollback history became full."""
        return self.__line_info[2]

    @property
    def first_visible_line_number(self) -> int:
        """
        Returns the line number of the first line currently displayed
        onscreen. Changes when the user scrolls."""
        return self.__line_info[3]


class Session:
    """
    Represents an iTerm2 session.
    """

    class Delegate:
        """
        Provides callbacks for Session.
        """
        @abc.abstractmethod
        def session_delegate_get_tab(
                self, session) -> typing.Optional['iterm2.Tab']:
            """Returns the tab for a session."""

        @abc.abstractmethod
        def session_delegate_get_window(
                self, session) -> typing.Optional['iterm2.Window']:
            """Returns the window for a session."""

        @abc.abstractmethod
        async def session_delegate_create_session(
                self, session_id: str) -> typing.Optional['iterm2.session.Session']:
            """Creates a new Session object given a session ID."""

    delegate: typing.Optional[Delegate] = None

    @staticmethod
    def active_proxy(connection: iterm2.connection.Connection) -> 'Session':
        """
        Use this to register notifications against the currently active
        session.

        :param connection: The connection to iTerm2.

        :returns: A proxy for the currently active session.
        """
        return ProxySession(connection, "active")

    @staticmethod
    def all_proxy(connection: iterm2.connection.Connection):
        """
        Use this to register notifications against all sessions, including
        those not yet created.

        :param connection: The connection to iTerm2.

        :returns: A proxy for all sessions.
        """
        return ProxySession(connection, "all")

    def __init__(self, connection, link, summary=None):
        """
        Do not call this yourself. Use :class:`~iterm2.app.App` instead.

        :param connection: :class:`Connection`
        :param link: :class:`iterm2.api_pb2.SplitTreeNode.SplitTreeLink`
        :param summary: :class:`iterm2.api_pb2.SessionSummary`
        """
        self.connection = connection

        if link is not None:
            self.__session_id = link.session.unique_identifier
            self.frame = link.session.frame
            self.__grid_size = link.session.grid_size
            self.name = link.session.title
            self.buried = False
        elif summary is not None:
            self.__session_id = summary.unique_identifier
            self.name = summary.title
            self.buried = True
            self.__grid_size = None
            self.frame = None
        self.__preferred_size = self.grid_size

    def __repr__(self):
        return "<Session name=%s id=%s>" % (self.name, self.__session_id)

    def to_session_summary_protobuf(self):
        """Returns the protobuf representation."""
        # pylint: disable=no-member
        summary = iterm2.api_pb2.SessionSummary()
        summary.unique_identifier = self.session_id
        summary.grid_size.width = self.preferred_size.width
        summary.grid_size.height = self.preferred_size.height
        return summary

    def update_from(self, session):
        """Replace internal state with that of another session."""
        self.frame = session.frame
        self.__grid_size = session.grid_size
        self.name = session.name

    def pretty_str(self, indent: str = "") -> str:
        """
        :returns: A string describing the session.
        """
        return indent + "Session \"%s\" id=%s %s frame=%s\n" % (
            self.name,
            self.__session_id,
            iterm2.util.size_str(self.grid_size),
            iterm2.util.frame_str(self.frame))

    @property
    def tab(self) -> typing.Optional['iterm2.Tab']:
        """Returns the containing tab."""
        # Note: App sets get_tab on Session when it's created.
        return Session.get_tab(self)

    @staticmethod
    def get_tab(obj: 'iterm2.Session') -> typing.Optional['iterm2.Tab']:
        """Returns the tab containing a given session.

        Do not call this before creating a :class:`~iterm2.app.App` object.
        """
        if Session.delegate:
            return Session.delegate.session_delegate_get_tab(obj)
        return None

    @staticmethod
    def get_window(obj: 'iterm2.Session') -> typing.Optional['iterm2.Window']:
        """Returns the window containing a given session.

        Do not call this before creating a :class:`~iterm2.app.App` object.
        """
        if Session.delegate:
            return Session.delegate.session_delegate_get_window(obj)
        return None

    @property
    def window(self) -> typing.Optional['iterm2.Window']:
        """Returns the containing terminal window."""
        return Session.get_window(self)

    @property
    def preferred_size(self) -> iterm2.util.Size:
        """
        The size in cells to resize to when `Tab.async_update_layout()` is
        called. The size is a :class:`iterm2.util.Size`.
        """
        return self.__preferred_size

    @preferred_size.setter
    def preferred_size(self, value: iterm2.util.Size):
        """
        Sets the size in cells to resize to when `Tab.async_update_layout()` is
        called. The size is a :class:`iterm2.util.Size`.
        """
        self.__preferred_size = value

    @property
    def session_id(self) -> str:
        """
        :returns: the globally unique identifier for this session.
        """
        return self.__session_id

    async def async_get_screen_contents(self) -> iterm2.screen.ScreenContents:
        """
        Returns the contents of the mutable area of the screen.

        :returns: A :class:`iterm2.screen.ScreenContents`, containing the
            screen contents.
        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.
        """
        # pylint: disable=no-member
        result = await iterm2.rpc.async_get_screen_contents(
            self.connection,
            self.session_id)
        if (result.get_buffer_response.status ==
                iterm2.api_pb2.GetBufferResponse.Status.Value("OK")):
            return iterm2.screen.ScreenContents(result.get_buffer_response)
        raise iterm2.rpc.RPCException(
            iterm2.api_pb2.GetBufferResponse.Status.Name(
                result.get_buffer_response.status))

    async def async_get_contents(
            self,
            first_line: int,
            number_of_lines: int) -> typing.List['iterm2.screen.LineContents']:
        """
        Returns the contents of the session within a given range of lines.

        Use `async_get_line_info` to determine the available line numbers.

        `first_lines` should be at least as large as `(await
        async_get_line_info()).overflow`.

        If it is less, then you won't get as many lines as you asked for.

        To use this reliably, you must call async_get_line_info() and this
        method in a :class:`~iterm2.transaction.Transaction` to ensure the
        session doesn't change between calls. See the example below.

        :param first_line: The first line number to fetch.
        :param number_of_lines: The number of lines to fetch.
        :returns: A list of :class:`iterm2.screen.LineContents`. If some were
            unavailable then a subset is returned.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.

        .. code-block:: python
          :caption: Example that prints the first ten lines of `session`.


          async with iterm2.Transaction(connection) as txn:
            li = await session.async_get_line_info()
            lines = await session.async_get_contents(li.overflow, 10)
          print(list(map(lambda line: line.string, lines)))

        """
        coord_range = iterm2.util.WindowedCoordRange(
            iterm2.util.CoordRange(
                iterm2.util.Point(0, first_line),
                iterm2.util.Point(0, first_line + number_of_lines)))

        response = await iterm2.rpc.async_get_screen_contents(
            self.connection,
            self.session_id,
            coord_range)
        # pylint: disable=no-member
        if (response.get_buffer_response.status ==
                iterm2.api_pb2.GetBufferResponse.Status.Value("OK")):
            contents = iterm2.screen.ScreenContents(
                response.get_buffer_response)
            result = []
            for i in range(contents.number_of_lines):
                result.append(contents.line(i))
            return result
        raise iterm2.rpc.RPCException(
            iterm2.api_pb2.GetBufferResponse.Status.Name(
                response.get_buffer_response.status))

    def get_screen_streamer(
            self, want_contents: bool = True) -> iterm2.screen.ScreenStreamer:
        """
        Provides a nice interface for receiving updates to the screen.

        The screen is the mutable part of a session (its last lines, excluding
        scrollback history).

        :param want_contents: If `True`, the screen contents will be provided.
            See :class:`~iterm2.screen.ScreenStreamer` for details.

        :returns: A new screen streamer, suitable for monitoring the contents
            of this session.

        .. code-block:: python
          :caption: Example that calls `do_something()` with screen contents as
              they arrive.

          async with session.get_screen_streamer() as streamer:
            while condition():
              contents = await streamer.async_get()
              do_something(contents)

        """
        return iterm2.screen.ScreenStreamer(
            self.connection,
            self.__session_id,
            want_contents=want_contents)

    async def async_send_text(
            self, text: str, suppress_broadcast: bool = False) -> None:
        """
        Send text as though the user had typed it.

        :param text: The text to send.
        :param suppress_broadcast: If `True`, text goes only to the specified
            session even if broadcasting is on.

        .. seealso::
            * Example ":ref:`broadcast_example`"
            * Example ":ref:`targeted_input_example`"
        """
        await iterm2.rpc.async_send_text(
            self.connection, self.__session_id, text, suppress_broadcast)

    async def async_split_pane(
            self,
            vertical: bool = False,
            before: bool = False,
            profile: typing.Union[None, str] = None,
            profile_customizations: typing.Union[
                None,
                iterm2.profile.LocalWriteOnlyProfile] = None) -> 'Session':
        """
        Splits the pane, creating a new session.

        :param vertical: If `True`, the divider is vertical, else horizontal.
        :param before: If `True`, the new session will be to the left of or
            above the session being split. Otherwise, it will be to the right
            of or below it.
        :param profile: The profile name to use. `None` for the default
            profile.
        :param profile_customizations: Changes to the profile that should
            affect only this session, or `None` to make no changes.

        :returns: A newly created Session.

        :throws: :class:`SplitPaneException` if something goes wrong.

        .. seealso:: Example ":ref:`broadcast_example`"
        """
        if profile_customizations is None:
            custom_dict = None
        else:
            custom_dict = profile_customizations.values

        result = await iterm2.rpc.async_split_pane(
            self.connection,
            self.__session_id,
            vertical,
            before,
            profile,
            profile_customizations=custom_dict)
        # pylint: disable=no-member
        if (result.split_pane_response.status ==
                iterm2.api_pb2.SplitPaneResponse.Status.Value("OK")):
            new_session_id = result.split_pane_response.session_id[0]
            assert Session.delegate
            session = await Session.delegate.session_delegate_create_session(
                new_session_id)
            if session:
                return session
            raise SplitPaneException(
                "No such session {}".format(new_session_id))
        # pylint: disable=no-member
        raise SplitPaneException(
            iterm2.api_pb2.SplitPaneResponse.Status.Name(
                result.split_pane_response.status))

    async def async_set_profile_properties(
            self,
            write_only_profile: iterm2.profile.LocalWriteOnlyProfile) -> None:
        """
        Sets the value of properties in this session.

        When you use this function the underlying profile is not modified. The
        session will keep a copy of its profile with these modifications.

        :param write_only_profile: A write-only profile that has the desired
            changes.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.

        .. seealso::
          * Example ":ref:`copycolor_example`"
          * Example ":ref:`settabcolor_example`"
          * Example ":ref:`increase_font_size_example`"
        """
        if iterm2.capabilities.supports_multiple_set_profile_properties(
                self.connection):
            assignments: typing.List[typing.Tuple[str, str]] = []
            for key, json_value in write_only_profile.values.items():
                assignments += [(key, json_value)]
            response = await iterm2.rpc.async_set_profile_properties_json(
                self.connection,
                self.session_id,
                assignments)
            status = response.set_profile_property_response.status
            # pylint: disable=no-member
            if (status != iterm2.api_pb2.SetProfilePropertyResponse.
                    Status.Value("OK")):
                raise iterm2.rpc.RPCException(
                    iterm2.api_pb2.SetProfilePropertyResponse.Status.Name(
                        status))
            return

        # Deprecated code path, in use by 3.3.0beta9 and earlier.
        for key, json_value in write_only_profile.values.items():
            response = await iterm2.rpc.async_set_profile_property_json(
                self.connection,
                self.session_id,
                key,
                json_value)
            status = response.set_profile_property_response.status
            # pylint: disable=no-member
            if (status != iterm2.api_pb2.SetProfilePropertyResponse.Status.
                    Value("OK")):
                raise iterm2.rpc.RPCException(
                    iterm2.api_pb2.SetProfilePropertyResponse.Status.Name(
                        status))

    async def async_get_profile(self) -> iterm2.profile.Profile:
        """
        Fetches the profile of this session

        :returns: The profile for this session, including any session-local
            changes not in the underlying profile.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.

        .. seealso::
            * Example ":ref:`blending_example`"
            * Example ":ref:`colorhost_example`"
            * Example ":ref:`current_preset_example`"
            * Example ":ref:`random_color_example`"
        """
        response = await iterm2.rpc.async_get_profile(
            self.connection, self.__session_id)
        status = response.get_profile_property_response.status
        # pylint: disable=no-member
        if status == iterm2.api_pb2.GetProfilePropertyResponse.Status.Value(
                "OK"):
            return iterm2.profile.Profile(
                self.__session_id,
                self.connection,
                response.get_profile_property_response.properties)
        raise iterm2.rpc.RPCException(
            iterm2.api_pb2.GetProfilePropertyResponse.Status.Name(status))

    async def async_set_profile(self, profile: iterm2.profile.Profile):
        """
        Changes this session's profile.

        The profile may be an existing profile, an existing
        profile with modifications, or a previously uknown
        profile with a unique GUID.

        :param profile: The `~iterm2.profile.Profile` to use.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.
        """
        await self.async_set_profile_properties(profile.local_write_only_copy)

    async def async_inject(self, data: bytes) -> None:
        """
        Injects data as though it were program output.

        :param data: A byte array to inject.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.

        .. seealso:: Example ":ref:`cls_example`"
        """
        response = await iterm2.rpc.async_inject(
            self.connection, data, [self.__session_id])
        status = response.inject_response.status[0]
        # pylint: disable=no-member
        if status != iterm2.api_pb2.InjectResponse.Status.Value("OK"):
            raise iterm2.rpc.RPCException(
                iterm2.api_pb2.InjectResponse.Status.Name(status))

    async def async_activate(
            self,
            select_tab: bool = True,
            order_window_front: bool = True) -> None:
        """
        Makes the session the active session in its tab.

        :param select_tab: Whether the tab this session is in should be
            selected.
        :param order_window_front: Whether the window this session is in should
            be brought to the front and given keyboard focus.

        .. seealso:: Example ":ref:`broadcast_example`"
        """
        await iterm2.rpc.async_activate(
            self.connection,
            True,
            select_tab,
            order_window_front,
            session_id=self.__session_id)

    async def async_set_variable(self, name: str, value: typing.Any):
        """
        Sets a user-defined variable in the session.

        See the Scripting Fundamentals documentation for more information on
        user-defined variables.

        :param name: The variable's name. Must begin with "user."
        :param value: The new value to assign.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.

        .. seealso:: Example ":ref:`escindicator_example`"
        """
        result = await iterm2.rpc.async_variable(
            self.connection,
            self.__session_id,
            [(name, json.dumps(value))],
            [])
        status = result.variable_response.status
        # pylint: disable=no-member
        if status != iterm2.api_pb2.VariableResponse.Status.Value("OK"):
            raise iterm2.rpc.RPCException(
                iterm2.api_pb2.VariableResponse.Status.Name(status))

    async def async_get_variable(self, name: str) -> typing.Any:
        """
        Fetches a session variable.

        See the Scripting Fundamentals documentation for more information on
        variables.

        :param name: The variable's name.

        :returns: The variable's value or empty string if it is undefined.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.

        .. seealso:: Example ":ref:`colorhost_example`"
        """
        result = await iterm2.rpc.async_variable(
            self.connection, self.__session_id, [], [name])
        status = result.variable_response.status
        # pylint: disable=no-member
        if status != iterm2.api_pb2.VariableResponse.Status.Value("OK"):
            raise iterm2.rpc.RPCException(
                iterm2.api_pb2.VariableResponse.Status.Name(status))
        return json.loads(result.variable_response.values[0])

    async def async_restart(self, only_if_exited: bool = False) -> None:
        """
        Restarts a session.

        :param only_if_exited: When `True`, this will raise an exception if the
            session is still running. When `False`, a running session will be
            killed and restarted.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.
        """
        result = await iterm2.rpc.async_restart_session(
            self.connection, self.__session_id, only_if_exited)
        status = result.restart_session_response.status
        # pylint: disable=no-member
        if status != iterm2.api_pb2.RestartSessionResponse.Status.Value("OK"):
            raise iterm2.rpc.RPCException(
                iterm2.api_pb2.RestartSessionResponse.Status.Name(status))

    async def async_close(self, force: bool = False) -> None:
        """
        Closes the session.

        :param force: If `True`, the user will not be prompted for a
            confirmation.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.
        """
        result = await iterm2.rpc.async_close(
            self.connection, sessions=[self.__session_id], force=force)
        status = result.close_response.statuses[0]
        # pylint: disable=no-member
        if status != iterm2.api_pb2.CloseResponse.Status.Value("OK"):
            raise iterm2.rpc.RPCException(
                iterm2.api_pb2.CloseResponse.Status.Name(status))

    async def async_set_grid_size(self, size: iterm2.util.Size) -> None:
        """Sets the visible size of a session.

        Note: This is meant for tabs that contain a single pane. If split panes
        are present, use :func:`~iterm2.tab.Tab.async_update_layout` instead.

        :param size: The new size for the session, in cells.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.

        Note: This will fail on fullscreen windows."""
        await self._async_set_property("grid_size", size.json)

    @property
    def grid_size(self) -> iterm2.util.Size:
        """Returns the size of the visible part of the session in cells.

        :returns: The size of the visible part of the session in cells.
        """
        return self.__grid_size

    async def async_set_buried(self, buried: bool) -> None:
        """Buries or disinters a session.

        :param buried: If `True`, bury the session. If `False`, disinter it.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.
        """
        await self._async_set_property("buried", json.dumps(buried))

    async def _async_set_property(self, key, json_value):
        """Sets a property on this session.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.
        """
        response = await iterm2.rpc.async_set_property(
            self.connection, key, json_value, session_id=self.session_id)
        status = response.set_property_response.status
        # pylint: disable=no-member
        if status != iterm2.api_pb2.SetPropertyResponse.Status.Value("OK"):
            raise iterm2.rpc.RPCException(
                iterm2.api_pb2.SetPropertyResponse.Status.Name(status))
        return response

    async def async_get_selection(self) -> iterm2.selection.Selection:
        """
        :returns: The selected regions of this session. The selection will be
            empty if there is no selected text.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.

        .. seealso:: Example ":ref:`georges_title_example`"
        """
        response = await iterm2.rpc.async_get_selection(
            self.connection, self.session_id)
        status = response.selection_response.status
        # pylint: disable=no-member
        if status != iterm2.api_pb2.SelectionResponse.Status.Value("OK"):
            raise iterm2.rpc.RPCException(
                iterm2.api_pb2.SelectionResponse.Status.Name(status))
        subs = []
        for sub_proto in (response.selection_response.get_selection_response.
                          selection.sub_selections):
            start = iterm2.util.Point(
                sub_proto.windowed_coord_range.coord_range.start.x,
                sub_proto.windowed_coord_range.coord_range.start.y)
            end = iterm2.util.Point(
                sub_proto.windowed_coord_range.coord_range.end.x,
                sub_proto.windowed_coord_range.coord_range.end.y)
            coord_range = iterm2.util.CoordRange(start, end)
            column_range = iterm2.util.Range(
                sub_proto.windowed_coord_range.columns.location,
                sub_proto.windowed_coord_range.columns.length)
            windowed_coord_range = iterm2.util.WindowedCoordRange(
                coord_range, column_range)

            sub = iterm2.SubSelection(
                windowed_coord_range,
                iterm2.selection.SelectionMode.from_proto_value(
                    sub_proto.selection_mode),
                sub_proto.connected)
            subs.append(sub)
        return iterm2.Selection(subs)

    async def async_get_selection_text(
            self, selection: iterm2.selection.Selection) -> str:
        """Fetches the text within a selection region.

        :param selection: A :class:`~iterm2.selection.Selection` defining a
            region in the session.

        .. seealso::
            * :func:`async_get_selection`.
            * Example ":ref:`georges_title_example`"

        :returns: A string with the selection's contents. Discontiguous
            selections are combined with newlines."""
        return await selection.async_get_string(
            self.connection,
            self.session_id,
            self.grid_size.width)

    async def async_set_selection(
            self, selection: iterm2.selection.Selection) -> None:
        """
        :param selection: The regions of text to select.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.

        .. seealso:: Example ":ref:`zoom_on_screen_example`"
        """
        response = await iterm2.rpc.async_set_selection(
            self.connection, self.session_id, selection)
        status = response.selection_response.status
        # pylint: disable=no-member
        if status != iterm2.api_pb2.SelectionResponse.Status.Value("OK"):
            raise iterm2.rpc.RPCException(
                iterm2.api_pb2.SelectionResponse.Status.Name(status))

    async def async_get_line_info(self) -> SessionLineInfo:
        """
        Fetches the number of lines that are visible, in history, and that have
        been removed after history became full.

        :returns: Information about the session's wrapped lines of text.

        .. seealso:: Example ":ref:`zoom_on_screen_example`"
        """
        response = await iterm2.rpc.async_get_property(
            self.connection,
            "number_of_lines",
            session_id=self.session_id)
        status = response.get_property_response.status
        # pylint: disable=no-member
        if status != iterm2.api_pb2.GetPropertyResponse.Status.Value("OK"):
            raise iterm2.rpc.RPCException(
                iterm2.api_pb2.GetPropertyResponse.Status.Name(status))
        dictionary = json.loads(response.get_property_response.json_value)
        values = (dictionary["grid"],
                  dictionary["history"],
                  dictionary["overflow"],
                  dictionary["first_visible"])
        return SessionLineInfo(values)

    async def async_set_name(self, name: str):
        """Changes the session's name.

        This is equivalent to editing the session's name manually in the Edit
        Session window.

        :param name: The new name to use.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.
        """
        invocation = iterm2.util.invocation_string(
            "iterm2.set_name",
            {"name": name})
        await iterm2.rpc.async_invoke_method(
            self.connection, self.session_id, invocation, -1)

    async def async_run_tmux_command(
            self, command: str, timeout: float = -1) -> str:
        """Invoke a tmux command and return its result. Raises an exception if
        this session is not a tmux integration session.

        :param command: The tmux command to run.
        :param timeout: The amount of time to wait for a response, or -1 to use
            the default.

        :returns: The output from tmux.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.
        """
        invocation = iterm2.util.invocation_string(
            "iterm2.run_tmux_command",
            {"command": command})
        return await iterm2.rpc.async_invoke_method(
            self.connection, self.session_id, invocation, timeout)

    async def async_invoke_function(
            self, invocation: str, timeout: float = -1):
        """
        Invoke an RPC. Could be a function registered by this or another
        script, or a built-in function.

        This invokes the RPC in the context of this session. Most user-defined
        RPCs are invoked in a session context (for example, invocations
        attached to triggers or key bindings). Default variables will be pulled
        from that scope. If you call a function from the wrong context it may
        fail because its defaults will not be set properly.

        :param invocation: A function invocation string.
        :param timeout: Max number of secondsto wait. Negative values mean to
            use the system default timeout.

        :returns: The result of the invocation if successful.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.
        """
        response = await iterm2.rpc.async_invoke_function(
            self.connection,
            invocation,
            session_id=self.session_id,
            timeout=timeout)
        which = response.invoke_function_response.WhichOneof('disposition')
        # pylint: disable=no-member
        if which == 'error':
            if (response.invoke_function_response.error.status ==
                    iterm2.api_pb2.InvokeFunctionResponse.Status.
                    Value("TIMEOUT")):
                raise iterm2.rpc.RPCException("Timeout")
            raise iterm2.rpc.RPCException("{}: {}".format(
                iterm2.api_pb2.InvokeFunctionResponse.Status.Name(
                    response.invoke_function_response.error.status),
                response.invoke_function_response.error.error_reason))
        return json.loads(
            response.invoke_function_response.success.json_result)

    async def async_get_coprocess(self) -> typing.Optional[str]:
        """
        Returns the command line of the currently running coprocess, if any.

        :returns: Whether there is a coprocess currently running.
        """
        iterm2.capabilities.check_supports_coprocesses(self.connection)
        invocation = iterm2.util.invocation_string(
            "iterm2.get_coprocess",
            {})
        return await iterm2.rpc.async_invoke_method(
            self.connection, self.session_id, invocation, -1)

    async def async_stop_coprocess(self) -> bool:
        """
        Stops the currently running coprocess, if any.

        :returns: True if a coprocess was stopped or False if non was running.
        """
        iterm2.capabilities.check_supports_coprocesses(self.connection)
        invocation = iterm2.util.invocation_string(
            "iterm2.stop_coprocess",
            {})
        return bool(
            await iterm2.rpc.async_invoke_method(
                self.connection, self.session_id, invocation, -1))

    async def async_add_annotation(self, range: iterm2.util.CoordRange, text: str):
        """
        Adds an annotation.

        This will not replace an existing annotation. It will always add a new one.

        Throws an exception if the range is invalid.

        :param range: The range to annotate.
        :param text: The string to annotate with.
        """
        iterm2.capabilities.check_supports_add_annotation(self.connection)
        args = {
            "startX": range.start.x,
            "startY": range.start.y,
            "endX": range.end.x,
            "endY": range.end.y,
            "text": text}
        invocation = iterm2.util.invocation_string("iterm2.add_annotation", args)
        await iterm2.rpc.async_invoke_method(self.connection, self.session_id, invocation, -1)

    async def async_run_coprocess(self, command_line: str) -> bool:
        """
        Runs a coprocess, provided non is already running.

        :param command_line: The command line for the new coprocess.

        :returns: True if it was launched or False if one was already running.
        """
        iterm2.capabilities.check_supports_coprocesses(self.connection)
        invocation = iterm2.util.invocation_string(
            "iterm2.run_coprocess",
            {"commandLine": command_line})
        return bool(
            await iterm2.rpc.async_invoke_method(
                self.connection, self.session_id, invocation, -1))


class InvalidSessionId(Exception):
    """The specified session ID is not allowed in this method."""


class ProxySession(Session):
    """A proxy for a Session.

    This is used when you specify an abstract session ID like "all" or
    "active".  Since the session or set of sessions that refers to is
    ever-changing, this proxy stands in for the real thing. It may limit
    functionality since it doesn't make sense to, for example, get the screen
    contents of "all" sessions.
    """
    def __init__(self, connection, session_id):
        super().__init__(connection, session_id)
        self.__session_id = session_id

    def __repr__(self):
        return "<ProxySession %s>" % self.__session_id

    def pretty_str(self, indent=""):
        return indent + "ProxySession %s" % self.__session_id
